/*
    Copyright (C) 2004 Dejvuth Suwimonteerabuth

    This file is part of jMoped.

    jMoped is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    jMoped is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with jMoped; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package jmoped;

import org.gjt.jclasslib.io.*;
import org.gjt.jclasslib.structures.*;
import org.gjt.jclasslib.structures.attributes.*;
import org.gjt.jclasslib.structures.constants.*;
import org.gjt.jclasslib.bytecode.*;

import java.util.ArrayList;
import java.io.File;

/**
 * Class information for the pushdown system.
 */
public class PDSClass {
	
	private String className;
	private String superClassName;
	private ArrayList includedClassNameList;
	private ArrayList includedPDSBasicMethodList;
	private PDSGlobal pdsGlobal;
	private MethodInfo[] methodInfo;
	private ArrayList pdsMethodList;
	private ArrayList newClassNameList;
	private int instanceId;
	private int initCount;
	private int initUsed;
	private boolean assertIncluded;
	
	/**
	 * Initializes a new class.
	 *
	 * @param classFile its ClassFile information.
	 * @param bits the number of bits of variables.
	 */
	public PDSClass(ClassFile classFile, int bits) throws Exception {
		this(classFile, null, bits);
	}
	
	/**
	 * Initializes a new class.
	 *
	 * @param classFile its ClassFile information.
	 * @param includedPDSBasicMethodList the list indicating methods that have 
	 *                                   already been included.
	 * @param bits the number of bits of variables.
	 */
	public PDSClass(ClassFile classFile, 
									ArrayList includedPDSBasicMethodList,
									int bits) throws Exception {
		className = classFile.getThisClassName().replace('/', '_');
		
		if (PDSConstant.verbose > 0) {
			System.out.println("\nClass: " + className);
		}
		
		superClassName = classFile.getSuperClassName();
		if (superClassName != null) {
			superClassName = superClassName.replace('/', '_');
			if (PDSConstant.verbose > 0) {
				System.out.println("Super Class: " + superClassName);
			}
		}
		else if (PDSConstant.verbose > 0) {
			System.out.println("Super Class is null");
		}
		this.includedPDSBasicMethodList = includedPDSBasicMethodList;
		pdsGlobal = new PDSGlobal(classFile, bits);
		assertIncluded = pdsGlobal.isAssert();
		methodInfo = classFile.getMethods();
		initPDSMethodList();
		initFields(classFile);
	}
	
	/**
	 * Returns this class name.
	 *
	 * @return this class name.
	 */
	public String getClassName() {
		return className;
	}
	
	/**
	 * Returns its super class name.
	 *
	 * @return its super class name.
	 */
	public String getSuperClassName() {
		return superClassName;
	}
	
	/**
	 * Returns information about global variables of this class.
	 *
	 * @return global variables' information.
	 */
	public PDSGlobal getPDSGlobal() {
		return pdsGlobal;
	}
	
	/**
	 * Returns the list of names of classes that have already been included.
	 *
	 * @return the list of included class.
	 */
	public ArrayList getIncludedClassNameList() {
		return includedClassNameList;
	}
	
	/**
	 * Returns the basic information about methods that have already been 
	 * included.
	 *
	 * @return the list of basic information about methods.
	 */
	public ArrayList getIncludedPDSBasicMethodList() {
		return includedPDSBasicMethodList;
	}
	
	/**
	 * Returns the list of names of classes, which is instantiated by this class.
	 *
	 * @return the class name list.
	 */
	public ArrayList getNewClassNameList() {
		return newClassNameList;
	}
	
	/**
	 * Returns the method name of the given index.
	 *
	 * @param index the method index.
	 * @return the method name.
	 */
	public String getMethodName(int index) {
		return getPDSMethod(index).getMethodName();
	}
	
	/**
	 * Returns the method index of the given name.
	 *
	 * @param methodName the method name.
	 * @return the method index.
	 */
	public int getMethodIndex(String methodName) {
		for (int i = 0; i < getMethodCount(); i++) {
			//System.out.println("      " + methodName + " " + getMethodName(i));
			if (getMethodName(i).equals(methodName))
				return i;
		}
		return -1;
	}
	
	public PDSMethod getPDSMethod(int index) {
		return ((PDSMethod) pdsMethodList.get(index));
	}
	
	public PDSMethod getPDSMethod(String methodName) {
		int methodIndex = getMethodIndex(methodName);
		if (methodIndex == -1) {
			return null;
		}
		return ((PDSMethod) pdsMethodList.get(methodIndex));
	}
	
	public PDSMethod getPDSMethodWithStackSymbol(String stackSymbol) {
		for (int i = 0; i < getMethodCount(); i++) {
			if (stackSymbol.startsWith(getMethodName(i))) {
				return getPDSMethod(i);
			}
		}
		return null;
	}
	
	/**
	 * Returns the number of methods in this class.
	 *
	 * @return the number of methods.
	 */
	public int getMethodCount() {
		return pdsMethodList.size();
	}
	
	/**
	 * Returns the list of bytecode instructions of the given method index.
	 *
	 * @param index the method index.
	 * @return the bytecode instructions list
	 */
	public ArrayList getInstructionList(int index) {
		return ((PDSMethod) pdsMethodList.get(index)).getInstructionList();
	}
	
	/**
	 * Returns the list of stack symbols of the given method index.
	 *
	 * @param index the method index.
	 * @return the stack symbols list.
	 */
	public ExtendedList getStackSymbolList(int index) {
		return ((PDSMethod) pdsMethodList.get(index)).getStackSymbolList();
	}
	
	/**
	 * Returns the stack symbol of the given method index and bytecode 
	 * instruction index.
	 *
	 * @param methodIndex the method index.
	 * @param instructionIndex the bytecode instruction index.
	 * @return the stack symbol.
	 */
	public String getStackSymbol(int methodIndex, int instructionIndex) {
		return ((PDSMethod) pdsMethodList.get(methodIndex)).getStackSymbol(
			instructionIndex);
	}
	
	/**
	 * Add the stack symbol into the list of the given method index.
	 *
	 * @param index the method index.
	 * @param stackSymbol the stack symbol to be added.
	 */
	public void addStackSymbol(int index, String stackSymbol) {
		((PDSMethod) pdsMethodList.get(index)).addStackSymbolList(stackSymbol);
	}
	
	/**
	 * Returns the maximum stack depth used by the method given by the index.
	 *
	 * @param index the method index.
	 * @return the maximum stack depth.
	 */
	public int getMaxStack(int index) {
		return ((PDSMethod) pdsMethodList.get(index)).getMaxStack();
	}
	
	/**
	 * Returns the maximum number of local variables used by the method given 
	 * by the index.
	 *
	 * @param index the method index.
	 * @return the maximum number of local variables.
	 */
	public int getMaxLocal(int index) {
		return ((PDSMethod) pdsMethodList.get(index)).getMaxLocal();
	}
	
	/**
	 * Returns the instance id of this class.
	 * The instance id is the unique number assigned to an instance in the 
	 * Java source.
	 *
	 * @return the instance id.
	 */
	public int getInstanceId() {
		return instanceId;
	}
	
	/**
	 * Set the instance id of this class.
	 * The instance id is the unique number assigned to an instance in the 
	 * Java source.
	 *
	 * @param instanceId the instance id.
	 */
	public void setInstanceId(int instanceId) {
		this.instanceId = instanceId;
	}
	
	/**
	 * Returns the number of instances having the type of this class.
	 *
	 * @return the number of instances.
	 */
	public int getInitCount() {
		return initCount;
	}
	
	/**
	 * Increments by one the number of instances having the type of this class.
	 */
	public void incInitCount() {
		initCount++;
	}
	
	/**
	 * Increment by the given value the number of instances having the type of 
	 * this class.
	 *
	 * @param i the value to be incremented.
	 */
	public void incInitCount(int i) {
		initCount += i;
	}
	
	/**
	 * Returns the number of instances having the type of this class that has 
	 * already been used.
	 *
	 * @return the number of instances used.
	 */
	public int getInitUsed() {
		return initUsed;
	}
	
	/**
	 * Increments the number of instances having the type of this class that 
	 * has already been used.
	 */
	public void incInitUsed() {
		initUsed++;
	}
	
	/**
	 * Returns the list of array sizes of the given method index.
	 *
	 * @param index the method index.
	 * @return array sizes list.
	 */
	public ArrayList getArraySizeList(int index) {
		return ((PDSMethod) pdsMethodList.get(index)).getArraySizeList();
	}
	
	/**
	 * Returns the array size of the given method and array index.
	 *
	 * @param methodIndex the method index.
	 * @param arraySizeIndex the array index.
	 * @return the array size.
	 */
	public int getArraySize(int methodIndex, int arraySizeIndex) {
		return ((PDSMethod) pdsMethodList.get(methodIndex)).getArraySize(
			arraySizeIndex);
	}
	
	/**
	 * Returns the number of arrays of the given method index.
	 *
	 * @param index the method index.
	 * @return the number of arrays.
	 */
	public int getArraySizeCount(int index) {
		return ((PDSMethod) pdsMethodList.get(index)).getArraySizeCount();
	}
	
	/**
	 * Tests if every method of this class returns nothing.
	 *
	 * @return <code>true</code> if every method returns void.
	 */
	public boolean isVoid() {
		for (int i = 0; i < getMethodCount(); i++)
			if (!isVoid(i))
				return false;
		return true;
	}
	
	/**
	 * Tests if the method at the index returns nothing.
	 *
	 * @param index the method index.
	 * @return <code>true</code> if the method returns void.
	 */
	public boolean isVoid(int index) {
		return getMethodName(index).charAt(getMethodName(index).length() - 1) 
			== 'V';
	}
	
	/**
	 * Tests if the method, given its name, returns nothing.
	 * The method name must follows the convention used in the pushdown system.
	 * That is, the last character of the method name indicates its return type.
	 *
	 * @param methodName the method name.
	 * @return <code>true</code> if the method returns void.
	 */
	public static boolean isVoid(String methodName) {
		return methodName.charAt(methodName.length() - 1) == 'V';
	}
	
	/**
	 * Returns the number of parameters of the method, given its index.
	 *
	 * @param index the method index.
	 * @return the number of parameters.
	 */
	public int getParameterCount(int index) {
		return ((PDSMethod) pdsMethodList.get(index)).getParamCount();
	}
	
	/**
	 * Returns the number of parameters given the method's descriptor in the 
	 * JVM standard.
	 *
	 * @param descriptor the method descriptor.
	 * @return the number of parameters.
	 */
	public static int getParameterCount(String descriptor) {
		int count = 0;
		String param = descriptor.substring(1, descriptor.indexOf(")"));
		
		for (int i = 0; i < param.length(); i++) {
			if (param.charAt(i) == '[') {
				continue;
			}
			if (param.charAt(i) == 'L') {
				while (param.charAt(i) != ';') {
					i++;
				}
			}
			count++;
		}
		
		return count;
		
		//String shortDescriptor = descriptor.replaceAll("[", "");
		//shortDescriptor = descriptor.replaceAll("L[a-zA-Z/]*;", "L");
		//return (shortDescriptor.indexOf(')') 
		//	- shortDescriptor.indexOf('(') - 1);
	}
	
	/**
	 * Checks if there is a method in this class that is the "class 
	 * initialization method" (the method having name <code>clinit</code>.)
	 *
	 * @return <code>true</code> if there is the class initialization method.
	 */
	public boolean isClinit() {
		String clinitName = className + PDSConstant.INIT_STATIC_METHOD_NAME;
		
		for (int i = 0; i < getMethodCount(); i++) {
			if (getMethodName(i).equals(clinitName)) {
				return true;
			}
		}
		return false;
	}
	
	/**
	 * Returns the offset of the bytecode instruction.
	 *
	 * @param methodIndex the method index.
	 * @param instructionIndex the instruction index.
	 * @return the offset.
	 */
	public int getOffset(int methodIndex, int instructionIndex) {
		return ((PDSMethod) pdsMethodList.get(methodIndex)).getOffset(
			instructionIndex).intValue();
	}
	
	/**
	 * Finds the corresponding program counter that is the handler when an 
	 * exeception occurs.
	 *
	 * @param methodIndex the method index.
	 * @param searchPc the program counter of interest.
	 * @return the handler program counter.
	 */
	public int getHandlerPc(int methodIndex, int searchPc) {
		ExceptionTableEntry[] exceptionTable 
			= ((PDSMethod) pdsMethodList.get(methodIndex)).getExceptionTable();
		for (int i = 0; i < exceptionTable.length; i++)
			if ((searchPc >= exceptionTable[i].getStartPc()) 
					&& (searchPc < exceptionTable[i].getEndPc()))
				return exceptionTable[i].getHandlerPc();
		return -1;
	}
	
	public boolean isAssert() {
		return assertIncluded;
	}
	
	private void initPDSMethodList() throws Exception {
		/*
		 * If there is no method to be included (from outside),
		 *   set pdsMethodList to every method in this class.
		 * Otherwise, for every to-be-included method belonging to this class
		 *   If this class contains the method, adds it to pdsMethodList
		 *   Otherwise, adds its superclass to includedClassNameList
		 *     removes the method from includedPDSBasicMethodList
		 *     changes the method name and adds it back.
		 */
		
		int i;
		int j;
		String methodName;
		ArrayList methodNameList = new ArrayList();
		ArrayList originalMethodNameList = new ArrayList();
		
		/* Initializes methodNameList */
		for (i = 0; i < methodInfo.length; i++) {
			methodName = getPDSMethodSignature(className + "_" 
				+ methodInfo[i].getName() + "_" 
				+ methodInfo[i].getDescriptor());
			methodNameList.add(methodName);
			originalMethodNameList.add(methodInfo[i].getName());
		}
		
		pdsMethodList = new ArrayList();
		includedClassNameList = new ArrayList();
		
		/* If no method to be included */
		if (includedPDSBasicMethodList == null) {
			for (i = 0; i < methodNameList.size(); i++) {
				if (((String) originalMethodNameList.get(i)).equals(
						PDSConstant.CLASS_DOLLARSIGN)) {
					continue;
				}
				pdsMethodList.add(new PDSMethod(className, 
					(String) methodNameList.get(i)));
			}
		}
		
		/* If there some methods to be included */
		else {
			PDSBasicMethod pdsBasicMethod;
			
			for (i = 0; i < includedPDSBasicMethodList.size(); i++) {
				pdsBasicMethod = (PDSBasicMethod) 
														includedPDSBasicMethodList.get(i);
				
				/*
				 * If this class contains a method to be included.
				 * Otherwise, do nothing.
				 */
				if (pdsBasicMethod.getClassName().equals(className)) {
					
					/* Finds the to-be-included method in this class. */
					for (j = 0; j < methodNameList.size(); j++) {
						if (pdsBasicMethod.getMethodName().equals(
								(String) methodNameList.get(j))) {
									
							if (PDSConstant.verbose > 2) {
								System.out.println("      className=" 
									+ className + ", methodName=" 
									+ (String) methodNameList.get(j));
							}
							
							pdsMethodList.add(new PDSMethod(className, 
								(String) methodNameList.get(j)));
							break;
						}
					}
					
					/* 
					 * If this method doesn't exist in methodInfo[], 
					 * it must exist in predecessor's class
					 */
					if (j == methodNameList.size()) {
						if (!isExistInIncludedClassNameList(superClassName) 
								&& !isNotToBeIncludedClass(superClassName)) {
							String changedMethodName;
							
							includedClassNameList.add(superClassName);
							pdsBasicMethod = (PDSBasicMethod) 
																	includedPDSBasicMethodList.remove(i);
							changedMethodName = superClassName 
								+ pdsBasicMethod.getMethodName().replaceFirst(className,
									"");
									
							if (PDSConstant.verbose > 2) { 
								System.out.println("changedMethodName: " 
									+ changedMethodName);
							}
							
							includedPDSBasicMethodList.add(
								new PDSMethod(superClassName, changedMethodName));
						}
					}
				}
			}
		}
		
		if (PDSConstant.verbose > 1) {
			System.out.println();
		}
	}
	
	private void initFields(ClassFile classFile) throws Exception {
		/* Initializes this fields */
		
		PDSMethod pdsMethod;
		AttributeInfo[] attributeInfo;
		CodeAttribute codeAttribute;
		AbstractInstruction inst;
		ImmediateShortInstruction immShortInst;
		ConstantClassInfo constantClassInfo;
		ConstantMethodrefInfo cpInfo;
		String calleeClassName;
		String calleeMethodName;
		String methodName;
		int methodIndex;
		int i;
		int j;
		int k;
		int l;
		
		if (includedPDSBasicMethodList == null)
			includedPDSBasicMethodList = new ArrayList();
		newClassNameList = new ArrayList();
		initCount = 0;
		initUsed = 0;
		
		for (methodIndex = 0; methodIndex < pdsMethodList.size(); 
				methodIndex++) {
			pdsMethod = (PDSMethod) pdsMethodList.get(methodIndex);
			
			/* Finds in methodInfo method having the same name as pdsMethod. */
			for (i = 0; i < methodInfo.length; i++) {
				methodName = getPDSMethodSignature(
					className + "_" + methodInfo[i].getName() + "_" 
					+ methodInfo[i].getDescriptor());
				if (pdsMethod.getMethodName().equals(methodName))
					break;
			}
			
			/* Breaks if cannot find the method. */
			if (i == methodInfo.length) {
				break;
			}
			
			if (PDSConstant.verbose > 1) {
				System.out.println("i = " + i 
					+ ", methodIndex = " + methodIndex 
					+ ", methodName = " + pdsMethod.getMethodName());
			}
			
			pdsMethod.setParamList(methodInfo[i].getDescriptor());
			attributeInfo = methodInfo[i].getAttributes();
			if(attributeInfo.length == 0)
				continue;
			
			for (j = 0; j < attributeInfo.length; j++) {
				
				/* If code attribute, otherwise do nothing. */
				if(attributeInfo[j].getName().equals("Code")) {
					codeAttribute = (CodeAttribute) attributeInfo[j];
					pdsMethod.setAccessFlag(methodInfo[i].getAccessFlags());
					pdsMethod.setMaxStack(codeAttribute.getMaxStack());
					pdsMethod.setMaxLocal(codeAttribute.getMaxLocals());
					pdsMethod.setInstructionList(
						(ArrayList) ByteCodeReader.readByteCode(
							codeAttribute.getCode()));
					pdsMethod.setExceptionTable(codeAttribute.getExceptionTable());
					
					/*
					 * For each instruction, there are 4 cases to consider:
					 * (1) if "new" opcode
					 * (2) if "invoke*" opcode
					 * (3) if "[a]newarray" opcode
					 * (4) if "multianewarray" opcode
					 */
					for (k = 0; k < pdsMethod.getInstructionCount(); k++) {
						inst = pdsMethod.getInstruction(k);
						
						if (PDSConstant.verbose > 2) {
							System.out.println("      " 
								+ inst.getOpcodeVerbose());
						}
						
						pdsMethod.addOffsetList(inst.getOffset());
						pdsMethod.addStackSymbolList(pdsMethod.getMethodName() 
							+ inst.getOffset());
						
						/* (1) If "new" opcode, adds new class to newClassNameList. */
						if (inst.getOpcode() == Opcodes.OPCODE_NEW) {
							immShortInst = (ImmediateShortInstruction) inst;
							constantClassInfo = (ConstantClassInfo) 
								classFile.getConstantPoolEntry(
									immShortInst.getImmediateShort(), 
									ConstantClassInfo.class);
									
							if (PDSConstant.verbose > 2) {
								System.out.println("new class name: " 
									+ constantClassInfo.getName().replace('/', '_'));
							}
							
							newClassNameList.add(
								constantClassInfo.getName().replace('/', '_'));
						}
						
						/* (2) If "invoke*" opcode. */
						if (isCallOpcode(inst.getOpcode())) {
							immShortInst = (ImmediateShortInstruction) 
								inst;
							cpInfo = (ConstantMethodrefInfo) 
								classFile.getConstantPoolEntry(
									immShortInst.getImmediateShort(), 
									ConstantMethodrefInfo.class);
							calleeClassName = cpInfo.getClassInfo().getName().replace(
								'/', '_');
							calleeMethodName = getPDSMethodSignature(calleeClassName + "_" 
								+ cpInfo.getNameAndTypeInfo().getVerbose());
							
							/*
							 * If this class is called class but the called method has not
							 * been included, adds the method to includedPDSBasicMethodList
							 * and redos all initializations again.
							 */
							if (calleeClassName.equals(className)
									&& !containsMethod(calleeMethodName)
									&& !cpInfo.getNameAndTypeInfo().getName().equals(
										PDSConstant.CLASS_DOLLARSIGN)) {
										
								if (PDSConstant.verbose > 1) { 
									System.out.println("   this new calleeMethodName: " 
										+ calleeMethodName);
								}
								
								includedPDSBasicMethodList.add(new PDSBasicMethod(className, 
									calleeMethodName));
								if (PDSConstant.verbose > 1) {
									System.out.println("   ***Recursive Call***");
									System.out.println();
								}
								initPDSMethodList();
								initFields(classFile);
								return;
							}
							
							/*
							 * If called class is not this class, adds class and method name
							 * to includedClassNameList and includedPDSBasicMethodList.
							 */
							if (!calleeClassName.equals(className)
									&& !isNotToBeIncludedClass(calleeClassName)
									&& !isNotToBeIncludedMethod(calleeMethodName)) {
								if (!isExistInIncludedClassNameList(calleeClassName)) {
									
									if (PDSConstant.verbose > 1) {
										System.out.println("   new calleeClassName: " 
											+ calleeClassName);
									}
									
									includedClassNameList.add(calleeClassName);
								}
								if (!isExistInIncludedPDSBasicMethodList(calleeMethodName)) {
									if (PDSConstant.verbose > 1) {
										System.out.println("   new calleeMethodName: " 
											+ calleeMethodName);
									}
									includedPDSBasicMethodList.add(
										new PDSBasicMethod(calleeClassName, calleeMethodName));
								}
							}
						}
						
						/* (3) If "[a]newarray" opcode. */
						if ((inst.getOpcode() == Opcodes.OPCODE_NEWARRAY) 
								|| (inst.getOpcode() 
									== Opcodes.OPCODE_ANEWARRAY)) {
							AbstractInstruction preInst  = pdsMethod.getInstruction(k - 1);
							int preOpcode = preInst.getOpcode();
							int size;
							if ((preOpcode >= Opcodes.OPCODE_ICONST_0) 
									&& (preOpcode <= Opcodes.OPCODE_ICONST_5)) {
								size = preOpcode - Opcodes.OPCODE_ICONST_0;
							} else if (preOpcode == Opcodes.OPCODE_BIPUSH) {
								size = ((ImmediateByteInstruction) 
									preInst).getImmediateByte();
							} else {
								size = 1;
							}
							
							if (pdsMethod.getMethodName().equals(
									className + PDSConstant.INIT_STATIC_METHOD_NAME)) {
								l = k;
								AbstractInstruction nextInst = pdsMethod.getInstruction(++l);
								while (nextInst.getOpcode()  != Opcodes.OPCODE_PUTSTATIC) {
									nextInst = pdsMethod.getInstruction(++l);
								}
								ImmediateShortInstruction nextImmShortInst 
									= (ImmediateShortInstruction) nextInst;
								String fieldName = PDSField.formatFieldName(
									classFile.getConstantPoolEntryName(
										nextImmShortInst.getImmediateShort()));
								pdsGlobal.setArraySize(fieldName, size);
							} else {
								pdsMethod.addArraySizeList(size);
							}
						}
						
						/* (4) If "multianewarray" opcode. */
						if (inst.getOpcode() == Opcodes.OPCODE_MULTIANEWARRAY) {
							int dim = ((MultianewarrayInstruction) inst).getDimensions();
							AbstractInstruction preInst;
							int preOpcode;
							int[] size = new int[dim];
							
							for (int m = 0; m < dim; m++) {
								preInst = pdsMethod.getInstruction(k-m-1);
								preOpcode = preInst.getOpcode();
								if ((preOpcode >= Opcodes.OPCODE_ICONST_0) 
										&& (preOpcode <= Opcodes.OPCODE_ICONST_5)) {
									size[m] = preOpcode - Opcodes.OPCODE_ICONST_0;
								} else if (preOpcode == Opcodes.OPCODE_BIPUSH) {
									size[m] = ((ImmediateByteInstruction) 
										preInst).getImmediateByte();
								} else
									size[m] = 1;
							}
							
							/* If this method is static init method */
							if (pdsMethod.getMethodName().equals(
									className + PDSConstant.INIT_STATIC_METHOD_NAME)) {
								l = k;
								AbstractInstruction nextAbstractInstruction 
									= pdsMethod.getInstruction(++l);
								while (nextAbstractInstruction.getOpcode() 
										!= Opcodes.OPCODE_PUTSTATIC)
									 nextAbstractInstruction = pdsMethod.getInstruction(++l);
								ImmediateShortInstruction nextImmediateShortInstruction 
									= (ImmediateShortInstruction) nextAbstractInstruction;
								String fieldName = PDSField.formatFieldName(
									classFile.getConstantPoolEntryName(
										nextImmediateShortInstruction.getImmediateShort()));
								
								int m = dim - 1;
								int outerLength = size[m];
								pdsGlobal.setArraySize(fieldName, outerLength);
								int count = 2;
								while (m > 0) {
									m--;
									for (int n = 0; n < outerLength; n++) {
										pdsGlobal.addArray(fieldName + "_" + count, className, 
											size[m]);
										count++;
									}
									outerLength *= size[m];
								}
							}
							else {
								int m = dim - 1;
								int outerLength = size[m];
								pdsMethod.addArraySizeList(outerLength);
								while (m > 0) {
									m--;
									for (int n = 0; n < outerLength; n++) {
										pdsMethod.addArraySizeList(size[m]);
									}
									outerLength *= size[m];
								}
							}
						}
					}
					break;
				}
			}
		}
		if (PDSConstant.verbose > 1) System.out.println();
	}
	
	private boolean isCallOpcode(int opcode) {
		if ((opcode == Opcodes.OPCODE_INVOKEVIRTUAL)
			|| (opcode == Opcodes.OPCODE_INVOKESPECIAL)
			|| (opcode == Opcodes.OPCODE_INVOKESTATIC))
			return true;
		else
			return false;
	}
	
	private boolean isNotToBeIncludedMethod(String methodName) {
		for (int i = 0; i < PDSConstant.NOT_TO_BE_INCLUDED_METHOD.length; i++)
			if (methodName.equals(PDSConstant.NOT_TO_BE_INCLUDED_METHOD[i]))
				return true;
		return false;
	}
	
	private boolean isNotToBeIncludedClass(String className) {
		for (int i = 0; i < PDSConstant.NOT_TO_BE_INCLUDED_CLASS.length; i++)
			if (className.equals(PDSConstant.NOT_TO_BE_INCLUDED_CLASS[i].replace(
					'/', '_')))
				return true;
		return false;
	}
	
	private boolean isExistInIncludedClassNameList(String className) {
		for (int i = 0; i < includedClassNameList.size(); i++)
			if (includedClassNameList.get(i).equals(className))
				return true;
		return false;
	}
	
	private boolean isExistInIncludedPDSBasicMethodList(String methodName) {
		for (int i = 0; i < includedPDSBasicMethodList.size(); i++) {
			if (((PDSBasicMethod) includedPDSBasicMethodList.get(i)).getMethodName(
					).equals(methodName))
				return true;
		}
		return false;
	}
	
	/**
	 * Checks if the given method name is in this class.
	 *
	 * @param methodName the method name.
	 * @return true, if the method is in this class.
	 */
	public boolean containsMethod(String methodName) {
		for (int i = 0; i < pdsMethodList.size(); i++) {
			if (((PDSMethod) pdsMethodList.get(i)).getMethodName().equals(
						methodName))
				return true;
		}
		return false;
	}
	
	private ArrayList getIncludedMethodNameList(String className) {
		ArrayList methodNameList = new ArrayList();
		
		PDSBasicMethod pdsBasicMethod;
		for (int i = 0; i < includedPDSBasicMethodList.size(); i++) {
			pdsBasicMethod = (PDSBasicMethod) includedPDSBasicMethodList.get(i);
			if (pdsBasicMethod.getClassName().equals(className))
				methodNameList.add(pdsBasicMethod.getMethodName());
		}
		return methodNameList;
	}
	
	/**
	 * Changes the descriptor format to the pushdown system style.
	 *
	 * @param descriptor the method's descriptor.
	 * @return the formatted method's descriptor.
	 */
	public static String getPDSMethodSignature(String descriptor) {
		String signature = descriptor;
		
		signature = signature.replaceAll("[()/;$]", "_");
		signature = signature.replace('[', '_');
		signature = signature.replace('.', '_');
		signature = signature.replaceAll("_{2,}", "_");
		signature = signature.replaceAll("[<>]", "_");
		if (signature.endsWith("_"))
			signature = signature.substring(0, signature.length()-1);
		
		return signature;
	}
	
}
