/*
    Copyright (C) 2004 Dejvuth Suwimonteerabuth

    This file is part of jMoped.

    jMoped is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    jMoped is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with jMoped; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package jmoped;

import org.gjt.jclasslib.structures.*;

/**
 * Field information for the pushdown system.
 */
public class PDSField {

	private String fieldName;
	private String className;
	private String type;
	private int accessFlag;
	private int fromIndex;
	private int toIndex;
	
	/**
	 * Initializes a static field with its name.
	 *
	 * @param fieldName the field name.
	 */
	public PDSField(String fieldName) {
		this(fieldName, null, AccessFlags.ACC_STATIC, 1, 1);
	}
	
	/**
	 * Initializes a static field with its name and its class name.
	 *
	 * @param fieldName the field name.
	 * @param className the class name.
	 */
	public PDSField(String fieldName, String className) {
		this(fieldName, className, AccessFlags.ACC_STATIC, 1, 1);
	}
	
	/**
	 * Initializes a field with its name, its class name, and its access flag.
	 *
	 * @param fieldName the field name.
	 * @param className the class name.
	 * @param accessFlag the access flag.
	 */
	public PDSField(String fieldName, String className, int accessFlag) {
		this(fieldName, className, accessFlag, 1, 1);
	}
	
	/**
	 * Initializes a field with its name, its class name, its access flag, 
	 * and indices. The indices indicate the instance ids having this field.
	 *
	 * @param fieldName the field name.
	 * @param className the class name.
	 * @param accessFlag the access flag.
	 * @param fromIndex from which instance id.
	 * @param toIndex to which instance id.
	 */
	public PDSField(String fieldName, 
									String className, 
									int accessFlag, 
									int fromIndex, 
									int toIndex) {
		this.fieldName = fieldName;
		this.className = className;
		this.type = "I";
		this.accessFlag = accessFlag;
		this.fromIndex = fromIndex;
		this.toIndex = toIndex;
	}
	
	/**
	 * Returns this field name.
	 *
	 * @return this field name.
	 */
	public String getFieldName() {
		return fieldName;
	}
	
	/**
	 * Returns the class name which this field belongs to.
	 *
	 * @return the class name.
	 */
	public String getClassName() {
		return className;
	}
	
	/**
	 * Set the type of this field
	 *
	 * @param type the type
	 */
	public void setType(String type) {
		this.type = type;
	}
	
	/**
	 * Returns this field type.
	 *
	 * @return this field type.
	 */
	public String getTypeVerbose() {
		
		if (!type.equals("I") && !type.equals("Z")) {
			System.err.println("Bug: unsupported type of field found");
			System.exit(1);
		}
		
		return "int";
	}
	
	/**
	 * Checks if this field is of boolean type
	 *
	 * @return <code>true</code> if boolean
	 */
	public boolean isBoolean() {
		return type.equals("Z");
	}
	
	/**
	 * Returns the lowest instance id that belongs to this field.
	 *
	 * @return the instance id.
	 */
	public int getFromIndex() {
		return fromIndex;
	}
	
	/**
	 * Set the lowest instance id that belongs to this field.
	 *
	 * @param fromIndex the instance id.
	 */
	public void setFromIndex(int fromIndex) {
		this.fromIndex = fromIndex;
	}
	
	/**
	 * Returns the highest instance id that belongs to this field.
	 *
	 * @return the instance id.
	 */
	public int getToIndex() {
		return toIndex;
	}
	
	/**
	 * Set the highest instance id that belongs to this field.
	 *
	 * @param toIndex the instance id.
	 */
	public void setToIndex(int toIndex) {
		this.toIndex = toIndex;
	}
	
	/**
	 * Increments the highest instance id that belongs to this field.
	 */
	public void incToIndex() {
		toIndex++;
	}
	
	/**
	 * Checks if this field is static.
	 *
	 * @return <code>true</code> if static.
	 */
	public boolean isStatic() {
		return (accessFlag & AccessFlags.ACC_STATIC) != 0;
	}
	
	/**
	 * Checks if the given access flag is static.
	 *
	 * @return <code>true</code> if static.
	 */
	public static boolean isStatic(int flag) {
		return (flag & AccessFlags.ACC_STATIC) != 0;
	}
	
	/**
	 * Formats the field name to the one legal to Moped.
	 *
	 * @param field the field name.
	 * @return the formatted field name.
	 */
	public static String formatFieldName(String field) {
		return field.replaceAll("[./$]", "_");
	}

}
