/*
    Copyright (C) 2004 Dejvuth Suwimonteerabuth

    This file is part of jMoped.

    jMoped is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    jMoped is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with jMoped; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package jmoped;

import org.gjt.jclasslib.structures.*;

import java.util.List;

/**
 * Information about global variables.
 */
public class PDSGlobal {
	
	private String type;
	private ExtendedList pdsFieldList;
	private ExtendedList pdsArrayList;
	private int bits;
	private boolean assertIncluded = false;
	
	/**
	 * Initializes without adding a variable.
	 *
	 * @param bits the number of bits of global variables.
	 */
	public PDSGlobal(int bits) throws Exception {
		type = "int";
		pdsFieldList = new ExtendedList();
		pdsArrayList = new ExtendedList();
		this.bits = bits;
	}
	
	/**
	 * Initializes global variables of the class.
	 *
	 * @param classFile the class file.
	 * @param bits the number of bits of global variables.
	 */
	public PDSGlobal(ClassFile classFile, int bits) throws Exception {
		type = "int";
		pdsFieldList = new ExtendedList();
		pdsArrayList = new ExtendedList();
		this.bits = bits;
		addGlobal(classFile);
	}
	
	/**
	 * Adds global variables of the class
	 *
	 * @param classFile the class file.
	 */
	public void addGlobal(ClassFile classFile) throws Exception {
		String className;
		String formattedFieldDescriptor;
		String fieldName;
		FieldInfo[] fieldInfo = classFile.getFields();
		PDSField pdsField;
		
		for (int i = 0; i < fieldInfo.length; i++) {
				
			if (fieldInfo[i].getName().equals(PDSConstant.ASSERTIONS_DISABLED)) {
				assertIncluded = true;
				continue;
			}
			if (fieldInfo[i].getName().startsWith(PDSConstant.CLASS_DOLLARSIGN)) {
				continue;
			}
			
			className = classFile.getThisClassName().replace('/', '_');
			fieldName = className + "_" + fieldInfo[i].getName().replace('$', '_');
			pdsField = new PDSField(fieldName, className, 
				fieldInfo[i].getAccessFlags());
			if (fieldInfo[i].getDescriptor().equals("Z")) {
				pdsField.setType("Z");
			}
			pdsFieldList.add(pdsField);
			
			if (fieldInfo[i].getDescriptor().indexOf('[') != -1) {
				if (PDSField.isStatic(fieldInfo[i].getAccessFlags())) {
					PDSArray pdsArray = new PDSArray(fieldName, className, 
						fieldInfo[i].getAccessFlags());
					if (fieldInfo[i].getDescriptor().indexOf('Z') != -1) {
						pdsArray.setType("Z");
					}
					pdsArrayList.add(pdsArray);
				}	
				else {
					System.out.println("Error: the current version does not " +
						"support non-static global array");
					System.exit(1);
				}
			}
			
		}
	}
	
	/**
	 * Adds the array field.
	 *
	 * @param fieldName the field name.
	 * @param className the class name of this field.
	 * @param size the size of the array.
	 */
	public void addArray(String fieldName, String className, int size) {
		PDSArray pdsArray = new PDSArray(fieldName, className);
		pdsArray.setArraySize(size);
		pdsArrayList.add(pdsArray);
	}
	
	/**
	 * Adds global variables from the other class.
	 *
	 * @param pdsGlobal the global variables information of the other class.
	 * @param instanceId the instance id of that class.
	 * @param initCount the number of instances of that class.
	 */
	public void addGlobal(PDSGlobal pdsGlobal, int instanceId, int initCount) {
		PDSField pdsField;
		List fieldList = pdsGlobal.getPDSFieldList();
		
		if (initCount == 0) {
			for (int i = 0; i < fieldList.size(); i++) {
				pdsField = (PDSField) fieldList.get(i);
				if (!pdsField.isStatic()) {
					fieldList.remove(i);
				}
			}	
		} else {
			for (int i = 0; i <  fieldList.size(); i++) {
				pdsField = (PDSField) fieldList.get(i);
				pdsField.setFromIndex(instanceId);
				pdsField.setToIndex(instanceId + initCount - 1);
			}
		}
		pdsFieldList.addAll(fieldList);
		pdsArrayList.addAll(pdsGlobal.getPDSArrayList());
		return;
	}
	
	/**
	 * Set the instance ids to all global fields.
	 *
	 * @param fromIndex from which instance id.
	 * @param toIndex to which instance id.
	 */
	public void setPDSFieldListIndex(int fromIndex, int toIndex) {
		PDSField pdsField;
		for (int i = 0; i < pdsFieldList.size(); i++) {
			pdsField = (PDSField) pdsFieldList.get(i);
			pdsField.setFromIndex(fromIndex);
			pdsField.setToIndex(toIndex);
		}
	}
	
	/**
	 * Defines the size of the global array.
	 *
	 * @param arrayName the array name.
	 * @param size the size of the array.
	 */
	public void setArraySize(String arrayName, int size) {
		PDSArray pdsArray;
		for (int i = 0; i < getArrayCount(); i++) {
			pdsArray = (PDSArray) pdsArrayList.get(i);
			if (pdsArray.getArrayName().equals(arrayName)) {
				pdsArray.setArraySize(size);
				break;
			}
		}
	}
	
	/**
	 * Adds the return variable to global variables.
	 */
	public void addReturnVariable() {
		pdsFieldList.add(new PDSField(PDSConstant.RETURN_VARIABLE));
	}
	
	/**
	 * Returns the list of global fields.
	 * Each component in the list is of type PDSField.
	 *
	 * @return the global field's list.
	 */
	public List getPDSFieldList() {
		return pdsFieldList;
	}
	
	/**
	 * Returns the list of global arrays.
	 * Each component in the list is of type PDSArray.
	 *
	 * @return the global array's list.
	 */
	public List getPDSArrayList() {
		return pdsArrayList;
	}
	
	/**
	 * Returns the size of the array.
	 *
	 * @param index the array index.
	 * @return the size of the array.
	 */
	public int getArraySize(int index) {
		return ((PDSArray) pdsArrayList.get(index)).getArraySize();
	}
	
	/**
	 * Returns the number of global fields
	 *
	 * @return the number of fields.
	 */
	public int getFieldCount() {
		return pdsFieldList.size();
	}
	
	/**
	 * Returns the number of global arrays.
	 *
	 * @return the number of arrays.
	 */
	public int getArrayCount() {
		return pdsArrayList.size();
	}
	
	/**
	 * Prints global variables declaration in Moped's syntax.
	 *
	 * @return global variables declaration.
	 */
	public String print() {
		StringBuffer output = new StringBuffer();
		
		output.append("global\n");
		output.append(pdsFieldList.printFieldList(bits));
		output.append(pdsArrayList.printArrayList(bits));
		
		return output.toString();
	}
	
	/**
	 * Prints global variables declaration in Romepla.
	 *
	 * @return global variables declaration.
	 */
	public String printRemopla() {
		StringBuffer output = new StringBuffer();
		
		output.append(pdsFieldList.printFieldList(bits));
		output.append(pdsArrayList.printArrayList(bits));
		
		return output.toString();
	}
	
	/**
	 * Returns the field information.
	 *
	 * @param index the field index.
	 * @return the field information.
	 */
	public PDSField get(int index) {
		return (PDSField) pdsFieldList.get(index);
	}
	
	/**
	 * Returns the array information.
	 *
	 * @param index the array index.
	 * @return the array information.
	 */
	public PDSArray getPDSArray(int index) {
		return (PDSArray) pdsArrayList.get(index);
	}
	
	/**
	 * Returns the number of global variables.
	 *
	 * @return the number of global variables.
	 */
	public int size() {
		return pdsFieldList.size() + pdsArrayList.size();
	}
	
	/**
	 * Tests if the given field name is already exist.
	 *
	 * @param fieldName the field name.
	 * @return <code>true</code> if the field is already exist.
	 */
	public boolean isInPDSFieldList(String fieldName) {
		for (int i = 0; i < pdsFieldList.size(); i++) {
			if (((PDSField) pdsFieldList.get(i)).getFieldName().equals(fieldName))
				return true;
		}
		return false;
	}
	
	public boolean isAssert() {
		return assertIncluded;
	}
	
}
