/*
    Copyright (C) 2004 Dejvuth Suwimonteerabuth

    This file is part of jMoped.

    jMoped is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    jMoped is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with jMoped; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package jmoped;

import org.gjt.jclasslib.io.*;
import org.gjt.jclasslib.structures.*;
import org.gjt.jclasslib.structures.attributes.*;
import org.gjt.jclasslib.structures.constants.*;
import org.gjt.jclasslib.bytecode.*;

import java.util.ArrayList;

/**
 * Method information for the pushdown system.
 */
public class PDSMethod extends PDSBasicMethod {
	
	private ArrayList instructionList;
	private ArrayList offsetList;
	private ExtendedList stackSymbolList;
	private ArrayList paramList;
	//private int paramCount;
	private int maxStack;
	private int maxLocal;
	private ArrayList arraySizeList;
	private ExceptionTableEntry[] exceptionTable;
	
	/**
	 * Initializes a new method.
	 *
	 * @param className the class name of this method.
	 * @param methodName this method name.
	 */
	public PDSMethod(String className, String methodName) {
		super(className, methodName);
		offsetList = new ArrayList();
		stackSymbolList = new ExtendedList();
		arraySizeList = new ArrayList();
	}
	
	/**
	 * Returns the list of bytecode instructions.
	 *
	 * @return bytecode instructions list.
	 */
	public ArrayList getInstructionList() {
		return instructionList;
	}
	
	/**
	 * Returns the bytecode instruction at the given index.
	 *
	 * @param index the bytecode instruction index.
	 * @return the bytecode instruction.
	 */
	public AbstractInstruction getInstruction(int index) {
		return (AbstractInstruction) instructionList.get(index);
	}
	
	/**
	 * Returns the number of bytecode instructions in this method.
	 *
	 * @return the number of bycode instructions.
	 */
	public int getInstructionCount() {
		return instructionList.size();
	}
	
	/**
	 * Returns the bytecode instruction offset.
	 *
	 * @param index the bytecode instruction index.
	 * @return the offset.
	 */
	public Integer getOffset(int index) {
		return (Integer) offsetList.get(index);
	}
	
	/**
	 * Returns the list of stack symbols of this method.
	 *
	 * @return stack symbols list.
	 */
	public ExtendedList getStackSymbolList() {
		return stackSymbolList;
	}
	
	/**
	 *  Returns the stack symbol at the given index.
	 *
	 * @param index the bytecode instruction index.
	 * @return the stack symbol.
	 */
	public String getStackSymbol(int index) {
		return (String) stackSymbolList.get(index);
	}
	
	
	/**
	 * Returns the list of parameters of this method.
	 *
	 * @return parameters list.
	 */
	public ArrayList getParamList() {
		return paramList;
	}
	
	/**
	 *  Returns the parameter at the given index.
	 *
	 * @param index the parameter index.
	 * @return the parameter.
	 */
	public String getParam(int index) {
		return (String) paramList.get(index);
	}
	
	
	/**
	 * Returns the number of parameters of this method.
	 *
	 * @return the number of paramters.
	 */
	public int getParamCount() {
		return paramList.size();
	}
	
	public boolean isParamArray(int index) {
		if (index > getParamCount() - 1) {
			return false;
		}
		return (getParam(index).charAt(0) == '[');
	}
	
	/**
	 * Returns the maximum stack depth used by this method.
	 *
	 * @return the maximum stack depth.
	 */
	public int getMaxStack() {
		return maxStack;
	}
	
	/**
	 * Returns the maximum number of local variables used by this method.
	 *
	 * @return the maximum number of local variables.
	 */
	public int getMaxLocal() {
		return maxLocal;
	}
	
	/**
	 * Returns the list of arrays used by this method.
	 *
	 * @return the array list.
	 */
	public ArrayList getArraySizeList() {
		return arraySizeList;
	}
	
	/**
	 * Returns the array size at the given index.
	 * @param index the array index.
	 * @return the array size.
	 */
	public int getArraySize(int index) {
		return ((Integer) arraySizeList.get(index)).intValue();
	}
	
	/**
	 * Counts the number of arrays in this method.
	 *
	 * @return the number of arrays.
	 */
	public int getArraySizeCount() {
		return arraySizeList.size();
	}
	
	/**
	 * Returns the exception table of this method.
	 *
	 * @return the exception table.
	 */
	public ExceptionTableEntry[] getExceptionTable() {
		return exceptionTable;
	}
	
	/**
	 * Finds the handler program counter if an exeption occurs.
	 *
	 * @param searchPc the program counter of interest.
	 * @return the handler program counter.
	 */
	public int getHandlerPc(int searchPc) {
		for (int i = 0; i < exceptionTable.length; i++)
			if ((searchPc >= exceptionTable[i].getStartPc()) && (searchPc < exceptionTable[i].getEndPc()))
				return exceptionTable[i].getHandlerPc();
		return -1;
	}
	
	/**
	 * Set the list of bytecode instructions of this method.
	 *
	 * @param instructionList the list of bytecode instructions.
	 */
	public void setInstructionList(ArrayList instructionList) {
		this.instructionList = instructionList;
	}
	
	/**
	 * Adds new instruction offset.
	 *
	 * @param offset the offset to be added.
	 */
	public void addOffsetList(int offset) {
		offsetList.add(new Integer(offset));
	}
	
	/**
	 * Adds new stack symbol.
	 *
	 * @param stackSymbol the stack symbol to be added.
	 */
	public void addStackSymbolList(String stackSymbol) {
		stackSymbolList.add(stackSymbol);
	}
	
	/**
	 * Set the list of parameters of this method.
	 *
	 * @param descriptor method descriptor in JVM's style.
	 */
	public void setParamList(String descriptor) {
		String param = descriptor.substring(1, descriptor.indexOf(")"));
		paramList = new ArrayList();
		char c;
		StringBuffer tmp;
		
		for (int i = 0; i < param.length(); i++) {
			switch (param.charAt(i)) {
				case 'I': 
					paramList.add("I");
					break;
					
				case 'Z': 
					paramList.add("Z");
					break;
					
				case '[': 
					tmp = new StringBuffer();
					while (param.charAt(i) == '[') {
						tmp.append("[");
						i++;
					}
					if (param.charAt(i) == 'L') {
						while (param.charAt(i) != ';') {
							tmp.append(param.charAt(i));
							i++;
						}
						tmp.append(";");
					} else {
						tmp.append(param.charAt(i));
					}
					paramList.add(tmp.toString());
					break;
					
				case 'L':
					tmp = new StringBuffer();
					while (param.charAt(i) != ';') {
						tmp.append(param.charAt(i));
						i++;
					}
					tmp.append(";");
					paramList.add(tmp.toString());
					break;
			}
		}
	}
	
	/**
	 * Set the number of parameters of this method.
	 *
	 * @param paramCount the number of parameters.
	 */
	//public void setParamCount(int paramCount) {
	//	this.paramCount = paramCount;
	//}
	
	/**
	 * Set the maximum stack depth of this method.
	 *
	 * @param maxStack the maximum stack depth.
	 */
	public void setMaxStack(int maxStack) {
		this.maxStack = maxStack;
	}
	
	/**
	 * Set the maximum local variables of this method.
	 *
	 * @param maxLocal the maximum local variables.
	 */
	public void setMaxLocal(int maxLocal) {
		this.maxLocal = maxLocal;
	}
	
	/**
	 * Adds new array.
	 *
	 * @param arraySize the size of the array to be added.
	 */
	public void addArraySizeList(int arraySize) {
		arraySizeList.add(new Integer(arraySize));
	}
	
	/**
	 * set the exception table of this method.
	 *
	 * @param exceptionTable the exception table.
	 */
	public void setExceptionTable(ExceptionTableEntry[] exceptionTable) {
		this.exceptionTable = exceptionTable;
	}
	
}
