/*
    Copyright (C) 2004 Dejvuth Suwimonteerabuth

    This file is part of jMoped.

    jMoped is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    jMoped is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with jMoped; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package jmoped;

import org.gjt.jclasslib.bytecode.*;
import org.gjt.jclasslib.structures.*;
import org.gjt.jclasslib.structures.constants.*;
import org.gjt.jclasslib.structures.attributes.*;

import java.util.ArrayList;
import java.util.List;

/**
 * List of rules of the pushdown system.
 */
public class PDSRuleList {
	
	private ClassFile classFile;
	private PDSClass pdsClass;
	private ArrayList pdsClassList;
	private PDSGlobal pdsGlobal;
	private ArrayList pdsRuleList;
	private ExtendedList pdsCond;
	private boolean assertIncluded;
	
	/**
	 * Initializes the rule list.
	 *
	 * @param classFile the class file.
	 * @param pdsClass the class information
	 * @param pdsClassList the list of all classes.
	 * @param pdsGlobal the global variables information.
	 */
	public PDSRuleList(ClassFile classFile, 
										 PDSClass pdsClass, 
										 ArrayList pdsClassList, 
										 PDSGlobal pdsGlobal) throws Exception {
		this.classFile = classFile;
		this.pdsClass = pdsClass;
		this.pdsClassList = pdsClassList;
		this.pdsGlobal = pdsGlobal;
		assertIncluded = false;
		initPDSRuleList();
	}
	
	/**
	 * Returns the size of this rule list.
	 *
	 * @return the rule list's size.
	 */
	public int size() {
		return pdsRuleList.size();
	}
	
	/**
	 * Adds another rule list to this rule list.
	 *
	 * @param anotherPDSRuleList the rule list to be added.
	 */
	public void addPDSRuleList(PDSRuleList anotherPDSRuleList) {
		pdsRuleList.addAll(anotherPDSRuleList.getPDSRuleList());
	}
	
	/**
	 * Returns the list of rules.
	 *
	 * @return the rule list.
	 */
	public ArrayList getPDSRuleList() {
		return pdsRuleList;
	}
	
	/**
	 * Returns the rule at the given index.
	 *
	 * @param index the rule index.
	 * @return the rule information.
	 */
	public PDSRule get(int index) {
		return (PDSRule) pdsRuleList.get(index);
	}
	
	private void initPDSRuleList() throws Exception {
		ArrayList instructionList;
		AbstractInstruction abstractInstruction;
		String[] stackCond;
		ArrayList returnAddressList;
		int newarrayCount;
		
		if (PDSConstant.verbose > 1) {
			System.out.println("initPDSRuleList(): " + pdsClass.getClassName());
		}
		
		pdsRuleList = new ArrayList();
		for (int methodIndex = 0; methodIndex < pdsClass.getMethodCount(); 
				methodIndex++) {
					
			if (PDSConstant.verbose > 0) {
				System.out.println("   methodIndex " + methodIndex + ": " 
					+ pdsClass.getMethodName(methodIndex));
			}
			
			returnAddressList = new ArrayList();
			newarrayCount = 0;
			instructionList = pdsClass.getInstructionList(methodIndex);
			
			if (instructionList == null) {
				continue;
			}
			
			for (int instructionIndex = 0; instructionIndex < instructionList.size(); 
					instructionIndex++) {
				
				abstractInstruction = (AbstractInstruction) instructionList.get(
					instructionIndex);
					
				if (PDSConstant.verbose > 2) {
					System.out.println("      " 
						+ abstractInstruction.getOpcodeVerbose());
				}
				
				switch (abstractInstruction.getOpcode()) {
					case Opcodes.OPCODE_NOP: 
						ruleDefault(abstractInstruction, methodIndex, instructionIndex);
						break;
					case Opcodes.OPCODE_ACONST_NULL: 
						ruleConst(abstractInstruction, methodIndex, instructionIndex, 
							Opcodes.OPCODE_ACONST_NULL);
						break;
					case Opcodes.OPCODE_ICONST_M1: 
						rulePush(abstractInstruction, methodIndex, instructionIndex);
						break;
					case Opcodes.OPCODE_ICONST_0: 
					case Opcodes.OPCODE_ICONST_1: 
					case Opcodes.OPCODE_ICONST_2: 
					case Opcodes.OPCODE_ICONST_3: 
					case Opcodes.OPCODE_ICONST_4: 
					case Opcodes.OPCODE_ICONST_5: 
						ruleConst(abstractInstruction, methodIndex, instructionIndex, 
							Opcodes.OPCODE_ICONST_0);
						break;
					case Opcodes.OPCODE_LCONST_0: 
					case Opcodes.OPCODE_LCONST_1: 
						ruleConst(abstractInstruction, methodIndex, instructionIndex, 
							Opcodes.OPCODE_LCONST_0);
						break;
					case Opcodes.OPCODE_FCONST_0: 
					case Opcodes.OPCODE_FCONST_1: 
					case Opcodes.OPCODE_FCONST_2: 
						ruleConst(abstractInstruction, methodIndex, instructionIndex, 
							Opcodes.OPCODE_FCONST_0);
						break;
					case Opcodes.OPCODE_DCONST_0: 
					case Opcodes.OPCODE_DCONST_1: 
						ruleConst(abstractInstruction, methodIndex, instructionIndex, 
							Opcodes.OPCODE_DCONST_0);
						break;
					case Opcodes.OPCODE_BIPUSH: 
						ruleBipush(abstractInstruction, methodIndex, instructionIndex);
						break;
					case Opcodes.OPCODE_SIPUSH: 
						ruleSipush(abstractInstruction, methodIndex, instructionIndex);
						break;
					case Opcodes.OPCODE_LDC: 
					case Opcodes.OPCODE_LDC_W: 
					case Opcodes.OPCODE_LDC2_W: 
						rulePush(abstractInstruction, methodIndex, instructionIndex);
						break;
					case Opcodes.OPCODE_ILOAD: 
					case Opcodes.OPCODE_LLOAD: 
						ruleLoad(abstractInstruction, methodIndex, instructionIndex);
						break;
					case Opcodes.OPCODE_FLOAD: 
					case Opcodes.OPCODE_DLOAD: 
						rulePush(abstractInstruction, methodIndex, instructionIndex);
						break;
					case Opcodes.OPCODE_ALOAD: 
						ruleLoad(abstractInstruction, methodIndex, instructionIndex);
						break;
					case Opcodes.OPCODE_ILOAD_0: 
					case Opcodes.OPCODE_ILOAD_1: 
					case Opcodes.OPCODE_ILOAD_2: 
					case Opcodes.OPCODE_ILOAD_3: 
						ruleLoad_x(abstractInstruction, methodIndex, instructionIndex, 
							Opcodes.OPCODE_ILOAD_0);
						break;
					case Opcodes.OPCODE_LLOAD_0: 
					case Opcodes.OPCODE_LLOAD_1: 
					case Opcodes.OPCODE_LLOAD_2: 
					case Opcodes.OPCODE_LLOAD_3: 
						ruleLoad_x(abstractInstruction, methodIndex, instructionIndex, 
							Opcodes.OPCODE_LLOAD_0);
						break;
					case Opcodes.OPCODE_FLOAD_0: 
					case Opcodes.OPCODE_FLOAD_1: 
					case Opcodes.OPCODE_FLOAD_2: 
					case Opcodes.OPCODE_FLOAD_3: 
					case Opcodes.OPCODE_DLOAD_0: 
					case Opcodes.OPCODE_DLOAD_1: 
					case Opcodes.OPCODE_DLOAD_2: 
					case Opcodes.OPCODE_DLOAD_3: 
						rulePush(abstractInstruction, methodIndex, instructionIndex);
						break;
					case Opcodes.OPCODE_ALOAD_0: 
					case Opcodes.OPCODE_ALOAD_1: 
					case Opcodes.OPCODE_ALOAD_2: 
					case Opcodes.OPCODE_ALOAD_3: 
						ruleLoad_x(abstractInstruction, methodIndex, instructionIndex, 
							Opcodes.OPCODE_ALOAD_0);
						break;
					case Opcodes.OPCODE_IALOAD: 
					case Opcodes.OPCODE_LALOAD: 
						ruleIaload(abstractInstruction, methodIndex, instructionIndex);
						break;
					case Opcodes.OPCODE_FALOAD: 
					case Opcodes.OPCODE_DALOAD: 
						rulePop(abstractInstruction, methodIndex, instructionIndex, 2, 1);
						break;
					case Opcodes.OPCODE_AALOAD: 
					case Opcodes.OPCODE_BALOAD: 
					case Opcodes.OPCODE_CALOAD: 
					case Opcodes.OPCODE_SALOAD: 
						ruleIaload(abstractInstruction, methodIndex, instructionIndex);
						break;
					case Opcodes.OPCODE_ISTORE: 
					case Opcodes.OPCODE_LSTORE: 
						ruleStore(abstractInstruction, methodIndex, instructionIndex);
						break;
					case Opcodes.OPCODE_FSTORE: 
					case Opcodes.OPCODE_DSTORE: 
						rulePop(abstractInstruction, methodIndex, instructionIndex, 1, 1);
						break;
					case Opcodes.OPCODE_ASTORE: 
						ruleStore(abstractInstruction, methodIndex, instructionIndex);
						break;
					case Opcodes.OPCODE_ISTORE_0: 
					case Opcodes.OPCODE_ISTORE_1: 
					case Opcodes.OPCODE_ISTORE_2: 
					case Opcodes.OPCODE_ISTORE_3: 
						ruleStore_x(abstractInstruction, methodIndex, instructionIndex, 
							Opcodes.OPCODE_ISTORE_0);
						break;
					case Opcodes.OPCODE_LSTORE_0: 
					case Opcodes.OPCODE_LSTORE_1: 
					case Opcodes.OPCODE_LSTORE_2: 
					case Opcodes.OPCODE_LSTORE_3: 
						ruleStore_x(abstractInstruction, methodIndex, instructionIndex, 
							Opcodes.OPCODE_LSTORE_0);
						break;
					case Opcodes.OPCODE_FSTORE_0: 
					case Opcodes.OPCODE_FSTORE_1: 
					case Opcodes.OPCODE_FSTORE_2: 
					case Opcodes.OPCODE_FSTORE_3: 
					case Opcodes.OPCODE_DSTORE_0: 
					case Opcodes.OPCODE_DSTORE_1: 
					case Opcodes.OPCODE_DSTORE_2: 
					case Opcodes.OPCODE_DSTORE_3: 
						rulePop(abstractInstruction, methodIndex, instructionIndex, 1, 1);
						break;
					case Opcodes.OPCODE_ASTORE_0: 
					case Opcodes.OPCODE_ASTORE_1: 
					case Opcodes.OPCODE_ASTORE_2: 
					case Opcodes.OPCODE_ASTORE_3: 
						ruleStore_x(abstractInstruction, methodIndex, instructionIndex, 
							Opcodes.OPCODE_ASTORE_0);
						break;
					case Opcodes.OPCODE_IASTORE: 
					case Opcodes.OPCODE_LASTORE: 
						ruleIastore(abstractInstruction, methodIndex, instructionIndex);
						break;
					case Opcodes.OPCODE_FASTORE: 
					case Opcodes.OPCODE_DASTORE: 
						rulePop(abstractInstruction, methodIndex, instructionIndex, 3, 1);
						break;
					case Opcodes.OPCODE_AASTORE: 
					case Opcodes.OPCODE_BASTORE: 
					case Opcodes.OPCODE_CASTORE: 
					case Opcodes.OPCODE_SASTORE: 
						ruleIastore(abstractInstruction, methodIndex, instructionIndex);
						break;
					case Opcodes.OPCODE_POP: 
						rulePop(abstractInstruction, methodIndex, instructionIndex, 1, 1);
						break;
					case Opcodes.OPCODE_POP2: 
						rulePop(abstractInstruction, methodIndex, instructionIndex, 2, 2);
						break;
					case Opcodes.OPCODE_DUP: 
						stackCond = new String[1];
						stackCond[0] = "sv0'=sv0";
						ruleDup(abstractInstruction, methodIndex, instructionIndex, 
							stackCond);
						break;
					case Opcodes.OPCODE_DUP_X1: 
					case Opcodes.OPCODE_DUP_X2: 
						stackCond = new String[3];
						stackCond[0] = "sv0'=sv0";
						stackCond[1] = "sv1'=sv1";
						stackCond[2] = "sv2'=sv0";
						ruleDup(abstractInstruction, methodIndex, instructionIndex, 
							stackCond);
						break;
					case Opcodes.OPCODE_DUP2: 
						stackCond = new String[1];
						stackCond[0] = "sv0'=sv0";
						ruleDup(abstractInstruction, methodIndex, instructionIndex, 
							stackCond);
						break;
					case Opcodes.OPCODE_DUP2_X1: 
					case Opcodes.OPCODE_DUP2_X2: 
						stackCond = new String[3];
						stackCond[0] = "sv0'=sv0";
						stackCond[1] = "sv1'=sv1";
						stackCond[2] = "sv2'=sv0";
						ruleDup(abstractInstruction, methodIndex, instructionIndex, 
							stackCond);
						break;
					case Opcodes.OPCODE_SWAP: 
						ruleSwap(abstractInstruction, methodIndex, instructionIndex);
						break;
					case Opcodes.OPCODE_IADD: 
					case Opcodes.OPCODE_LADD: 
						ruleTwoOperandExpression(abstractInstruction, methodIndex, 
							instructionIndex, "sv0'=sv1+sv0");
						break;
					case Opcodes.OPCODE_FADD: 
					case Opcodes.OPCODE_DADD: 
						rulePop(abstractInstruction, methodIndex, instructionIndex, 2, 1);
						break;
					case Opcodes.OPCODE_ISUB: 
					case Opcodes.OPCODE_LSUB: 
						ruleTwoOperandExpression(abstractInstruction, methodIndex, 
							instructionIndex, "sv0'=sv1-sv0");
						break;
					case Opcodes.OPCODE_FSUB: 
					case Opcodes.OPCODE_DSUB: 
						rulePop(abstractInstruction, methodIndex, instructionIndex, 2, 1);
						break;
					case Opcodes.OPCODE_IMUL: 
					case Opcodes.OPCODE_LMUL: 
						ruleTwoOperandExpression(abstractInstruction, methodIndex, 
							instructionIndex, "sv0'=sv1*sv0");
						break;
					case Opcodes.OPCODE_FMUL: 
					case Opcodes.OPCODE_DMUL: 
						rulePop(abstractInstruction, methodIndex, instructionIndex, 2, 1);
						break;
					case Opcodes.OPCODE_IDIV: 
					case Opcodes.OPCODE_LDIV: 
						ruleTwoOperandExpression(abstractInstruction, methodIndex, 
							instructionIndex, "sv0'=sv1/sv0");
						break;
					case Opcodes.OPCODE_FDIV: 
					case Opcodes.OPCODE_DDIV: 
						rulePop(abstractInstruction, methodIndex, instructionIndex, 2, 1);
						break;
					case Opcodes.OPCODE_IREM: 
					case Opcodes.OPCODE_LREM: 
						ruleTwoOperandExpression(abstractInstruction, methodIndex, 
							instructionIndex, "sv0'=sv1-(sv0*(sv1/sv0))");
						break;
					case Opcodes.OPCODE_FREM: 
					case Opcodes.OPCODE_DREM: 
						rulePop(abstractInstruction, methodIndex, instructionIndex, 2, 1);
						break;
					case Opcodes.OPCODE_INEG: 
					case Opcodes.OPCODE_LNEG: 
						ruleOneOperandExpression(abstractInstruction, methodIndex, 
							instructionIndex, "sv0'=0-sv0");
						break;
					case Opcodes.OPCODE_FNEG: 
					case Opcodes.OPCODE_DNEG: 
						ruleDefault(abstractInstruction, methodIndex, instructionIndex, 1);
						break;
					case Opcodes.OPCODE_ISHL: 
					case Opcodes.OPCODE_LSHL: 
						ruleTwoOperandExpression(abstractInstruction, methodIndex, 
							instructionIndex, "sv0'=sv1<<sv0");
						break;
					case Opcodes.OPCODE_ISHR: 
					case Opcodes.OPCODE_LSHR: 
					case Opcodes.OPCODE_IUSHR: 
					case Opcodes.OPCODE_LUSHR: 
					case Opcodes.OPCODE_IAND: 
					case Opcodes.OPCODE_LAND: 
					case Opcodes.OPCODE_IOR: 
					case Opcodes.OPCODE_LOR: 
					case Opcodes.OPCODE_IXOR: 
					case Opcodes.OPCODE_LXOR: 
						rulePop(abstractInstruction, methodIndex, instructionIndex, 2, 1);
						break;
					case Opcodes.OPCODE_IINC: 
						ruleIinc(abstractInstruction, methodIndex, instructionIndex);
						break;
					case Opcodes.OPCODE_LCMP: 
						ruleLcmp(abstractInstruction, methodIndex, instructionIndex);
						break;
					case Opcodes.OPCODE_FCMPL: 
					case Opcodes.OPCODE_FCMPG: 
					case Opcodes.OPCODE_DCMPL: 
					case Opcodes.OPCODE_DCMPG: 
						rulePop(abstractInstruction, methodIndex, instructionIndex, 2, 1);
						break;
					case Opcodes.OPCODE_IFEQ: 
						ruleIf(abstractInstruction, methodIndex, instructionIndex, 
							"sv0=0", "sv0!=0");
						break;
					case Opcodes.OPCODE_IFNE: 
						ruleIf(abstractInstruction, methodIndex, instructionIndex, 
							"sv0!=0", "sv0=0");
						break;
					case Opcodes.OPCODE_IFLT: 
						ruleIf(abstractInstruction, methodIndex, instructionIndex, 
							"sv0<0", "sv0>=0");
						break;
					case Opcodes.OPCODE_IFGE: 
						ruleIf(abstractInstruction, methodIndex, instructionIndex, 
							"sv0>=0", "sv0<0");
						break;
					case Opcodes.OPCODE_IFGT: 
						ruleIf(abstractInstruction, methodIndex, instructionIndex, 
							"sv0>0", "sv0<=0");
						break;
					case Opcodes.OPCODE_IFLE: 
						ruleIf(abstractInstruction, methodIndex, instructionIndex, 
							"sv0<=0", "sv0>0");
						break;
					case Opcodes.OPCODE_IF_ICMPEQ: 
						ruleIf_icmp(abstractInstruction, methodIndex, instructionIndex, 
							"sv1=sv0", "sv1!=sv0");
						break;
					case Opcodes.OPCODE_IF_ICMPNE: 
						ruleIf_icmp(abstractInstruction, methodIndex, instructionIndex, 
							"sv1!=sv0", "sv1=sv0");
						break;
					case Opcodes.OPCODE_IF_ICMPLT: 
						ruleIf_icmp(abstractInstruction, methodIndex, instructionIndex, 
							"sv1<sv0", "sv1>=sv0");
						break;
					case Opcodes.OPCODE_IF_ICMPGE: 
						ruleIf_icmp(abstractInstruction, methodIndex, instructionIndex, 
							"sv1>=sv0", "sv1<sv0");
						break;
					case Opcodes.OPCODE_IF_ICMPGT: 
						ruleIf_icmp(abstractInstruction, methodIndex, instructionIndex, 
							"sv1>sv0", "sv1<=sv0");
						break;
					case Opcodes.OPCODE_IF_ICMPLE: 
						ruleIf_icmp(abstractInstruction, methodIndex, instructionIndex, 
							"sv1<=sv0", "sv1>sv0");
						break;
					case Opcodes.OPCODE_IF_ACMPEQ: 
						ruleIf_icmp(abstractInstruction, methodIndex, instructionIndex, 
							"sv1=sv0", "sv1!=sv0");
						break;
					case Opcodes.OPCODE_IF_ACMPNE: 
						ruleIf_icmp(abstractInstruction, methodIndex, instructionIndex, 
							"sv1!=sv0", "sv1=sv0");
						break;
					case Opcodes.OPCODE_GOTO: 
						ruleGoto(abstractInstruction, methodIndex, instructionIndex);
						break;
					case Opcodes.OPCODE_JSR: 
						returnAddressList.add(new Integer(pdsClass.getOffset(methodIndex, 
							instructionIndex + 1)));
						ruleJsr(abstractInstruction, methodIndex, instructionIndex);
						break;
					case Opcodes.OPCODE_RET: 
						ruleRet(abstractInstruction, methodIndex, instructionIndex, 
							returnAddressList);
						break;
					case Opcodes.OPCODE_TABLESWITCH: 
						ruleTableswitch(abstractInstruction, methodIndex, instructionIndex);
						break;
					case Opcodes.OPCODE_LOOKUPSWITCH: 
						ruleLookupswitch(abstractInstruction, methodIndex, 
							instructionIndex);
						break;
					case Opcodes.OPCODE_IRETURN: 
					case Opcodes.OPCODE_LRETURN: 
					case Opcodes.OPCODE_FRETURN: 
					case Opcodes.OPCODE_DRETURN: 
					case Opcodes.OPCODE_ARETURN: 
						ruleIreturn(abstractInstruction, methodIndex, instructionIndex);
						break;
					case Opcodes.OPCODE_RETURN: 
						ruleReturn(abstractInstruction, methodIndex, instructionIndex);
						break;
					case Opcodes.OPCODE_GETSTATIC: 
						ruleGetstatic(abstractInstruction, methodIndex, instructionIndex);
						break;
					case Opcodes.OPCODE_PUTSTATIC: 
						rulePutstatic(abstractInstruction, methodIndex, instructionIndex);
						break;
					case Opcodes.OPCODE_GETFIELD: 
						ruleGetfield(abstractInstruction, methodIndex, instructionIndex);
						break;
					case Opcodes.OPCODE_PUTFIELD: 
						rulePutfield(abstractInstruction, methodIndex, instructionIndex);
						break;
					case Opcodes.OPCODE_INVOKEVIRTUAL: 
						ruleInvokevirtual(abstractInstruction, methodIndex, 
							instructionIndex);
						break;
					case Opcodes.OPCODE_INVOKESPECIAL: 
						ruleInvokespecial(abstractInstruction, methodIndex, 
							instructionIndex);
						break;
					case Opcodes.OPCODE_INVOKESTATIC: 
						ruleInvokestatic(abstractInstruction, methodIndex, 
							instructionIndex);
						break;
					case Opcodes.OPCODE_INVOKEINTERFACE: 
						ruleInvokeinterface(abstractInstruction, methodIndex, 
							instructionIndex);
						break;
					case Opcodes.OPCODE_NEW: 
						ruleNew(abstractInstruction, methodIndex, instructionIndex);
						break;
					case Opcodes.OPCODE_NEWARRAY: 
					case Opcodes.OPCODE_ANEWARRAY: 
						newarrayCount++;
						ruleNewarray(abstractInstruction, methodIndex, instructionIndex, 
							newarrayCount);
						break;
					case Opcodes.OPCODE_ARRAYLENGTH: 
						ruleArraylength(abstractInstruction, methodIndex, instructionIndex);
						break;
					case Opcodes.OPCODE_ATHROW: 
						ruleAthrow(abstractInstruction, methodIndex, instructionIndex);
						break;
					case Opcodes.OPCODE_CHECKCAST: 
						ruleCheckcast(abstractInstruction, methodIndex, instructionIndex);
						break;
					case Opcodes.OPCODE_INSTANCEOF: 
						ruleInstanceof(abstractInstruction, methodIndex, instructionIndex);
						break;
					case Opcodes.OPCODE_MULTIANEWARRAY: 
						newarrayCount++;
						newarrayCount = ruleMultianewarray(abstractInstruction, 
							methodIndex, instructionIndex, newarrayCount);
						break;
					case Opcodes.OPCODE_IFNULL: 
						ruleIf(abstractInstruction, methodIndex, instructionIndex, 
							"sv0=0", "sv0!=0");
						break;
					case Opcodes.OPCODE_IFNONNULL: 
						ruleIf(abstractInstruction, methodIndex, instructionIndex, 
							"sv0!=0", "sv0=0");
						break;	
					default:
						ruleDefault(abstractInstruction, methodIndex, instructionIndex);
						break;
				}
			}
		}
		
		if (assertIncluded) {
			PDSConfig pdsConfigFrom;
			PDSConfig pdsConfigTo;
			StringBuffer pdsLabel = new StringBuffer();
			
			pdsCond = new ExtendedList();
			pdsConfigFrom = new PDSConfig(PDSConstant.ASSERTION_ERROR);
			pdsConfigTo = new PDSConfig(PDSConstant.ASSERTION_ERROR);
			pdsLabel.append(Opcodes.OPCODE_GOTO);
			addDefaultGlobalCond();
			pdsRuleList.add(new PDSRule(pdsConfigFrom, pdsConfigTo, 
				pdsLabel.toString(), pdsCond));
		}
		
		if (PDSConstant.verbose > 0) {
			System.out.println();
		}
	}
	
	private void ruleConst(AbstractInstruction inst, 
												 int methodIndex, 
												 int instIndex, 
												 int baseOpcode) {
		PDSConfig pdsConfigFrom;
		PDSConfig pdsConfigTo;
		
		pdsCond = new ExtendedList();
		pdsConfigFrom = new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instIndex));
		pdsConfigTo = new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instIndex + 1));
		
		addDefaultGlobalCond();
		addDefaultLocalCond(methodIndex);
		addPushStackCond(methodIndex, inst.getOpcode() - baseOpcode);
		pdsRuleList.add(new PDSRule(pdsConfigFrom, pdsConfigTo, 
			inst.getOpcode(), pdsCond));
	}
	
	private void ruleBipush(AbstractInstruction inst, 
													int methodIndex, 
													int instIndex) {
		PDSConfig pdsConfigFrom;
		PDSConfig pdsConfigTo;
		ImmediateByteInstruction immByteInst;
		StringBuffer pdsLabel = new StringBuffer();
		
		pdsCond = new ExtendedList();
		pdsConfigFrom = new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instIndex));
		pdsConfigTo = new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instIndex + 1));
		immByteInst = (ImmediateByteInstruction) inst;
		
		pdsLabel.append(immByteInst.getOpcode());
		pdsLabel.append(PDSConstant.ONE_SPACE);
		pdsLabel.append(immByteInst.getImmediateByte());
			
		addDefaultGlobalCond();
		addDefaultLocalCond(methodIndex);
		addPushStackCond(methodIndex, immByteInst.getImmediateByte());
		pdsRuleList.add(new PDSRule(pdsConfigFrom, pdsConfigTo, 
			pdsLabel.toString(), pdsCond));
	}
	
	private void ruleSipush(AbstractInstruction inst, 
													int methodIndex, 
													int instIndex) {
		PDSConfig pdsConfigFrom;
		PDSConfig pdsConfigTo;
		ImmediateShortInstruction immShortInst;
		StringBuffer pdsLabel = new StringBuffer();
		
		pdsCond = new ExtendedList();
		pdsConfigFrom = new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instIndex));
		pdsConfigTo = new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instIndex + 1));
		immShortInst = (ImmediateShortInstruction) inst;
		
		pdsLabel.append(immShortInst.getOpcode());
		pdsLabel.append(PDSConstant.ONE_SPACE);
		pdsLabel.append(immShortInst.getImmediateShort());
		
		addDefaultGlobalCond();
		addDefaultLocalCond(methodIndex);
		addPushStackCond(methodIndex, immShortInst.getImmediateShort());
		pdsRuleList.add(new PDSRule(pdsConfigFrom, pdsConfigTo, 
			pdsLabel.toString(), pdsCond));
	}
	
	private void rulePush(AbstractInstruction inst, 
												int methodIndex, 
												int instIndex) {
		rulePush(inst, methodIndex, instIndex, "0");
	}
	
	private void rulePush(AbstractInstruction inst, 
												int methodIndex, 
												int instIndex,
												String pdsLabel) {
		PDSConfig pdsConfigFrom;
		PDSConfig pdsConfigTo;
		
		pdsCond = new ExtendedList();
		pdsConfigFrom = new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instIndex));
		pdsConfigTo = new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instIndex + 1));
		addDefaultGlobalCond();
		addDefaultLocalCond(methodIndex);
		addFakePushStackCond(methodIndex);
		pdsRuleList.add(new PDSRule(pdsConfigFrom, pdsConfigTo, 
			pdsLabel, pdsCond));
	}
	
	private void ruleLoad(AbstractInstruction inst, 
												int methodIndex, 
												int instIndex) {
		PDSConfig pdsConfigFrom;
		PDSConfig pdsConfigTo;
		ImmediateByteInstruction immByteInst;
		StringBuffer pdsLabel = new StringBuffer();
													
		pdsCond = new ExtendedList();
		pdsConfigFrom = new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instIndex));
		pdsConfigTo = new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instIndex + 1));
		immByteInst = (ImmediateByteInstruction) inst;
		
		pdsLabel.append(immByteInst.getOpcode());
		pdsLabel.append(PDSConstant.ONE_SPACE);
		pdsLabel.append(immByteInst.getImmediateByte());
		
		addDefaultGlobalCond();
		addDefaultLocalCond(methodIndex);
		addPushStackCond(methodIndex, "var" + immByteInst.getImmediateByte());
		pdsRuleList.add(new PDSRule(pdsConfigFrom, pdsConfigTo, 
			pdsLabel.toString(), pdsCond));
	}
	
	private void ruleLoad_x(AbstractInstruction inst, 
													int methodIndex, 
													int instIndex, 
													int baseOpcode) {
		PDSConfig pdsConfigFrom;
		PDSConfig pdsConfigTo;
		
		pdsCond = new ExtendedList();
		pdsConfigFrom = new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instIndex));
		pdsConfigTo = new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instIndex + 1));
		addDefaultGlobalCond();
		addDefaultLocalCond(methodIndex);
		addPushStackCond(methodIndex, "var" 
			+ (inst.getOpcode() - baseOpcode));
		pdsRuleList.add(new PDSRule(pdsConfigFrom, pdsConfigTo, 
			inst.getOpcode(), pdsCond));
	}
	
	private void ruleIaload(AbstractInstruction inst, 
													int methodIndex, 
													int instIndex) {
		PDSConfig pdsConfigFrom;
		PDSConfig pdsConfigTo;
		
		pdsCond = new ExtendedList();
		pdsConfigFrom = new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instIndex));
		pdsConfigTo = new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instIndex + 1));
		addDefaultGlobalCond();
		addDefaultLocalCond(methodIndex);
		for (int i = 0; i < pdsGlobal.getArrayCount() 
				+ pdsClass.getArraySizeList(methodIndex).size(); i++) {
			pdsCond.add("!(sv1=" + String.valueOf(i+1) + ") | (sv0'=arr" 
				+ String.valueOf(i+1) + "[sv0])");
		}
		addPopStackCond(methodIndex, 2, 1, false);
		pdsRuleList.add(new PDSRule(pdsConfigFrom, pdsConfigTo, 
			inst.getOpcode(), pdsCond));
	}
	
	private void ruleStore(AbstractInstruction inst, 
												 int methodIndex, 
												 int instIndex) {
		PDSConfig pdsConfigFrom;
		PDSConfig pdsConfigTo;
		ImmediateByteInstruction immByteInst;
		StringBuffer pdsLabel = new StringBuffer();
		
		pdsCond = new ExtendedList();
		pdsConfigFrom = new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instIndex));
		pdsConfigTo = new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instIndex + 1));
		immByteInst = (ImmediateByteInstruction) inst;
		
		pdsLabel.append(immByteInst.getOpcode());
		pdsLabel.append(PDSConstant.ONE_SPACE);
		pdsLabel.append(immByteInst.getImmediateByte());
		
		addDefaultGlobalCond();
		addStoreLocalCond(methodIndex, immByteInst.getImmediateByte());
		addPopStackCond(methodIndex);
		pdsRuleList.add(new PDSRule(pdsConfigFrom, pdsConfigTo, 
			pdsLabel.toString(), pdsCond));
	}
	
	private void ruleStore_x(AbstractInstruction inst, 
													 int methodIndex, 
													 int instIndex, 
													 int baseOpcode) {
		PDSConfig pdsConfigFrom;
		PDSConfig pdsConfigTo;
		
		pdsCond = new ExtendedList();
		pdsConfigFrom = new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instIndex));
		pdsConfigTo = new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instIndex + 1));
		
		addDefaultGlobalCond();
		addStoreLocalCond(methodIndex, inst.getOpcode() - baseOpcode);
		addPopStackCond(methodIndex);
		pdsRuleList.add(new PDSRule(pdsConfigFrom, pdsConfigTo, 
			inst.getOpcode(), pdsCond));
	}
	
	private void ruleIastore(AbstractInstruction inst, 
													 int methodIndex, 
													 int instIndex) {
		PDSConfig pdsConfigFrom;
		PDSConfig pdsConfigTo;
		
		pdsCond = new ExtendedList();
		pdsConfigFrom = new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instIndex));
		pdsConfigTo = new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instIndex + 1));
		addStoreArrayCond(methodIndex);
		addPopStackCond(methodIndex, 3, 3, false);
		pdsRuleList.add(new PDSRule(pdsConfigFrom, pdsConfigTo, 
			inst.getOpcode(), pdsCond));
	}
	
	private void rulePop(AbstractInstruction inst, 
											 int methodIndex, 
											 int instIndex, 
											 int from, 
											 int step) {
		rulePop(inst, methodIndex, instIndex, from, step, "0");
	}
	
	private void rulePop(AbstractInstruction inst, 
											 int methodIndex, 
											 int instIndex, 
											 int from, 
											 int step, 
											 String pdsLabel) {
		PDSConfig pdsConfigFrom;
		PDSConfig pdsConfigTo;
		
		pdsCond = new ExtendedList();
		pdsConfigFrom = new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instIndex));
		pdsConfigTo = new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instIndex + 1));
		addDefaultGlobalCond();
		addDefaultLocalCond(methodIndex);
		addPopStackCond(methodIndex, from, step, false);
		pdsRuleList.add(new PDSRule(pdsConfigFrom, pdsConfigTo, pdsLabel, pdsCond));
	}
	
	private void ruleDup(AbstractInstruction inst, 
											 int methodIndex, 
											 int instIndex, 
											 String[] stackCond) {
		PDSConfig pdsConfigFrom;
		PDSConfig pdsConfigTo;
		
		pdsCond = new ExtendedList();
		pdsConfigFrom = new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instIndex));
		pdsConfigTo = new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instIndex + 1));
		addDefaultGlobalCond();
		addDefaultLocalCond(methodIndex);
		for (int i = 0; i < stackCond.length; i++) {
			pdsCond.add(stackCond[i]);
		}
		addFakePushStackCond(methodIndex, stackCond.length);
		pdsRuleList.add(new PDSRule(pdsConfigFrom, pdsConfigTo, 
			inst.getOpcode(), pdsCond));
	}
	
	private void ruleSwap(AbstractInstruction inst, 
												int methodIndex, 
												int instIndex) {
		PDSConfig pdsConfigFrom;
		PDSConfig pdsConfigTo;
		
		pdsCond = new ExtendedList();
		pdsConfigFrom = new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instIndex));
		pdsConfigTo = new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instIndex + 1));
		addDefaultGlobalCond();
		addDefaultLocalCond(methodIndex);
		pdsCond.add("sv0'=sv1");
		pdsCond.add("sv1'=sv0");
		addDefaultStackCond(methodIndex, 2, false);
		pdsRuleList.add(new PDSRule(pdsConfigFrom, pdsConfigTo, inst.getOpcode(), 
			pdsCond));
	}
	
	private void ruleTwoOperandExpression(AbstractInstruction inst, 
																				int methodIndex, 
																				int instIndex, 
																				String expression) {
		PDSConfig pdsConfigFrom;
		PDSConfig pdsConfigTo;
		
		pdsCond = new ExtendedList();
		pdsConfigFrom = new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instIndex));
		pdsConfigTo = new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instIndex + 1));
		addDefaultGlobalCond();
		addDefaultLocalCond(methodIndex);
		pdsCond.add(expression);
		addPopStackCond(methodIndex, 2, 1, false);
		pdsRuleList.add(new PDSRule(pdsConfigFrom, pdsConfigTo, 
			inst.getOpcode(), pdsCond));
	}
	
	private void ruleOneOperandExpression(AbstractInstruction inst, 
																				int methodIndex, 
																				int instIndex, 
																				String expression) {
		PDSConfig pdsConfigFrom;
		PDSConfig pdsConfigTo;
		
		pdsCond = new ExtendedList();
		pdsConfigFrom = new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instIndex));
		pdsConfigTo = new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instIndex + 1));
		addDefaultGlobalCond();
		addDefaultLocalCond(methodIndex);
		pdsCond.add(expression);
		addDefaultStackCond(methodIndex, 1, false);
		pdsRuleList.add(new PDSRule(pdsConfigFrom, pdsConfigTo, 
			inst.getOpcode(), pdsCond));
	}
	
	private void ruleIinc(AbstractInstruction inst, 
												int methodIndex, 
												int instIndex) {
		PDSConfig pdsConfigFrom;
		PDSConfig pdsConfigTo;
		IncrementInstruction incInst;
		ImmediateByteInstruction immByteInst;
		StringBuffer pdsLabel = new StringBuffer();
		
		pdsCond = new ExtendedList();
		pdsConfigFrom = new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instIndex));
		pdsConfigTo = new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instIndex + 1));
		incInst = (IncrementInstruction) inst;
		immByteInst = (ImmediateByteInstruction) inst;
		
		pdsLabel.append(inst.getOpcode());
		pdsLabel.append(PDSConstant.ONE_SPACE);
		pdsLabel.append(immByteInst.getImmediateByte());
		pdsLabel.append(PDSConstant.ONE_SPACE);
		pdsLabel.append(incInst.getIncrementConst());
		
		addDefaultGlobalCond();
		if (incInst.getIncrementConst() <= 127) {
			addStoreLocalCond(methodIndex, immByteInst.getImmediateByte()
				, "var" + immByteInst.getImmediateByte() + "+" 
				+ incInst.getIncrementConst());
		} else {
			addStoreLocalCond(methodIndex, immByteInst.getImmediateByte()
				, "var" + immByteInst.getImmediateByte() + "-" 
				+ String.valueOf(256 - incInst.getIncrementConst()));
		}
		addDefaultStackCond(methodIndex);
		pdsRuleList.add(new PDSRule(pdsConfigFrom, pdsConfigTo,
			pdsLabel.toString(), pdsCond));
	}
	
	private void ruleLcmp(AbstractInstruction inst, 
												int methodIndex, 
												int instIndex) {
		PDSConfig pdsConfigFrom;
		PDSConfig pdsConfigTo;
		
		pdsCond = new ExtendedList();
		pdsConfigFrom = new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instIndex));
		pdsConfigTo = new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instIndex + 1));
			
		addDefaultGlobalCond();
		addDefaultLocalCond(methodIndex);
		pdsCond.add("!(sv1>sv0) | (sv0'=1)");
		pdsCond.add("!(sv1=sv0) | (sv0'=0)");
		addPopStackCond(methodIndex, 2, 1, false);
		pdsRuleList.add(new PDSRule(pdsConfigFrom, pdsConfigTo,
			inst.getOpcode(), pdsCond));
	}
	
	private void ruleIf(AbstractInstruction inst, 
											int methodIndex, 
											int instIndex, 
											String cond1, 
											String cond2) throws Exception {
		PDSConfig pdsConfigFrom;
		PDSConfig pdsConfigTo1;
		PDSConfig pdsConfigTo2;
		BranchInstruction branchInst;
		int targetOffset;
		StringBuffer pdsLabel = new StringBuffer();
		
		pdsCond = new ExtendedList();
		branchInst = (BranchInstruction) inst;
		targetOffset = branchInst.getOffset() + branchInst.getBranchOffset();
		pdsConfigFrom = new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instIndex));
		pdsConfigTo1 = new PDSConfig(pdsClass.getMethodName(methodIndex) 
			+ String.valueOf(targetOffset));
		
		pdsLabel.append(branchInst.getOpcode());
		pdsLabel.append(PDSConstant.ONE_SPACE);
		pdsLabel.append(targetOffset);
		
		addDefaultGlobalCond();
		addDefaultLocalCond(methodIndex);
		if (cond1 != null) {
			pdsCond.add(cond1);
		}
		addPopStackCond(methodIndex);
		pdsRuleList.add(new PDSRule(pdsConfigFrom, pdsConfigTo1, 
			pdsLabel.toString(), pdsCond));
		
		pdsConfigTo2 = new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instIndex + 1));
		pdsCond = new ExtendedList();
		addDefaultGlobalCond();
		addDefaultLocalCond(methodIndex);
		if (cond2 != null)
			pdsCond.add(cond2);
		addPopStackCond(methodIndex);
		pdsRuleList.add(new PDSRule(pdsConfigFrom, pdsConfigTo2, 
			pdsLabel.toString(), pdsCond));
	}
	
	
	private void ruleIf_icmp(AbstractInstruction inst, 
													 int methodIndex, 
													 int instIndex, 
													 String cond1, 
													 String cond2) throws Exception {
		PDSConfig pdsConfigFrom;
		PDSConfig pdsConfigTo1;
		PDSConfig pdsConfigTo2;
		BranchInstruction branchInst;
		int targetOffset;
		StringBuffer pdsLabel = new StringBuffer();
		
		pdsCond = new ExtendedList();
		branchInst = (BranchInstruction) inst;
		targetOffset = branchInst.getOffset() + branchInst.getBranchOffset();
		pdsConfigFrom = new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instIndex));
		pdsConfigTo1 = new PDSConfig(pdsClass.getMethodName(methodIndex) 
			+ String.valueOf(targetOffset));
		
		pdsLabel.append(branchInst.getOpcode());
		pdsLabel.append(PDSConstant.ONE_SPACE);
		pdsLabel.append(targetOffset);
		
		addDefaultGlobalCond();
		addDefaultLocalCond(methodIndex);
		pdsCond.add(cond1);
		addPopStackCond(methodIndex, 2, 2, false);
		pdsRuleList.add(new PDSRule(pdsConfigFrom, pdsConfigTo1, 
			pdsLabel.toString(), pdsCond));
		
		pdsConfigTo2 = new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instIndex + 1));
		pdsCond = new ExtendedList();
		addDefaultGlobalCond();
		addDefaultLocalCond(methodIndex);
		pdsCond.add(cond2);
		addPopStackCond(methodIndex, 2, 2, false);
		pdsRuleList.add(new PDSRule(pdsConfigFrom, pdsConfigTo2, 
			pdsLabel.toString(), pdsCond));
	}
	
	
	private void ruleGoto(AbstractInstruction inst, 
												int methodIndex, 
												int instIndex) throws Exception {
		PDSConfig pdsConfigFrom;
		PDSConfig pdsConfigTo;
		BranchInstruction branchInst;
		int targetOffset;
		StringBuffer pdsLabel = new StringBuffer();
		
		pdsCond = new ExtendedList();
		branchInst = (BranchInstruction) inst;
		targetOffset = branchInst.getOffset() + branchInst.getBranchOffset();
		pdsConfigFrom = new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instIndex));
		pdsConfigTo = new PDSConfig(pdsClass.getMethodName(methodIndex) 
			+ String.valueOf(targetOffset));
			
		pdsLabel.append(branchInst.getOpcode());
		pdsLabel.append(PDSConstant.ONE_SPACE);
		pdsLabel.append(targetOffset);
			
		addDefaultGlobalCond();
		addDefaultLocalCond(methodIndex);
		addDefaultStackCond(methodIndex);
		pdsRuleList.add(new PDSRule(pdsConfigFrom, pdsConfigTo, 
			pdsLabel.toString(), pdsCond));
	}
	
	private void ruleJsr(AbstractInstruction inst, 
											 int methodIndex, 
											 int instIndex) throws Exception {
		PDSConfig pdsConfigFrom;
		PDSConfig pdsConfigTo;
		BranchInstruction branchInst;
		int targetOffset;
		StringBuffer pdsLabel = new StringBuffer();
		
		pdsCond = new ExtendedList();
		branchInst = (BranchInstruction) inst;
		targetOffset = branchInst.getOffset() + branchInst.getBranchOffset();
		pdsConfigFrom = new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instIndex));
		pdsConfigTo = new PDSConfig(pdsClass.getMethodName(methodIndex) 
			+ String.valueOf(targetOffset));
		
		pdsLabel.append(branchInst.getOpcode());
		pdsLabel.append(PDSConstant.ONE_SPACE);
		pdsLabel.append(targetOffset);
		
		addDefaultGlobalCond();
		addDefaultLocalCond(methodIndex);
		addPushStackCond(methodIndex, String.valueOf(pdsClass.getOffset(
			methodIndex, instIndex + 1)));
		pdsRuleList.add(new PDSRule(pdsConfigFrom, pdsConfigTo,
			pdsLabel.toString(), pdsCond));
	}
	
	private void ruleRet(AbstractInstruction inst, 
											 int methodIndex, 
											 int instIndex, 
											 List returnAddressList) throws Exception {
		PDSConfig pdsConfigFrom;
		PDSConfig pdsConfigTo;
		ImmediateByteInstruction immByteInst;
		StringBuffer pdsLabel;
		
		for (int i = 0; i < returnAddressList.size(); i++) {
			pdsCond = new ExtendedList();
			pdsConfigFrom = new PDSConfig(pdsClass.getStackSymbol(
				methodIndex, instIndex));
			pdsConfigTo = new PDSConfig(pdsClass.getMethodName(methodIndex) 
				+ ((Integer) returnAddressList.get(i)).toString());
			immByteInst = (ImmediateByteInstruction) inst;
			
			pdsLabel = new StringBuffer();
			pdsLabel.append(immByteInst.getOpcode());
			pdsLabel.append(PDSConstant.ONE_SPACE);
			pdsLabel.append(immByteInst.getImmediateByte());
				
			addDefaultGlobalCond();
			pdsCond.add("var" + immByteInst.getImmediateByte() 
				+ "=" + ((Integer) returnAddressList.get(i)).toString());
			addDefaultLocalCond(methodIndex);
			addDefaultStackCond(methodIndex);
			pdsRuleList.add(new PDSRule(pdsConfigFrom, pdsConfigTo, 
				pdsLabel.toString(), pdsCond));
		}
	}
	
	private void ruleTableswitch(AbstractInstruction inst, 
															 int methodIndex, 
															 int instIndex) throws Exception {
		PDSConfig pdsConfigFrom;
		PDSConfig pdsConfigTo;
		TableSwitchInstruction tableSwitchInst;
		int[] jumpOffsets;
		
		tableSwitchInst = (TableSwitchInstruction) inst;
		jumpOffsets = tableSwitchInst.getJumpOffsets();
		pdsConfigFrom = new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instIndex));
		for (int i = 0; 
				i <= (tableSwitchInst.getHighByte() - tableSwitchInst.getLowByte()); 
				i++) {
			pdsCond = new ExtendedList();
			pdsConfigTo = new PDSConfig(pdsClass.getMethodName(methodIndex) 
				+ String.valueOf(tableSwitchInst.getOffset() + jumpOffsets[i]));
			addDefaultGlobalCond();
			addDefaultLocalCond(methodIndex);
			pdsCond.add("sv0=" + String.valueOf(tableSwitchInst.getLowByte() + i));
			addPopStackCond(methodIndex);
			pdsRuleList.add(new PDSRule(pdsConfigFrom, pdsConfigTo, pdsCond));
		}
		pdsCond = new ExtendedList();
		pdsConfigTo = new PDSConfig(pdsClass.getMethodName(methodIndex) 
			+ String.valueOf(tableSwitchInst.getOffset() 
				+ tableSwitchInst.getDefaultOffset()));
		addDefaultGlobalCond();
		addDefaultLocalCond(methodIndex);
		for (int i = 0; 
				i <= (tableSwitchInst.getHighByte() - tableSwitchInst.getLowByte()); 
				i++) {
			pdsCond.add("sv0!=" + String.valueOf(tableSwitchInst.getLowByte() + i));
		}
		addPopStackCond(methodIndex);
		pdsRuleList.add(new PDSRule(pdsConfigFrom, pdsConfigTo, pdsCond));
	}
	
	private void ruleLookupswitch(AbstractInstruction inst, 
																int methodIndex, 
																int instIndex) throws Exception {
		PDSConfig pdsConfigFrom;
		PDSConfig pdsConfigTo;
		LookupSwitchInstruction lookupSwitchInst;
		List matchOffsetPairs;
		
		lookupSwitchInst = (LookupSwitchInstruction) inst;
		pdsConfigFrom =  new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instIndex));
		matchOffsetPairs = lookupSwitchInst.getMatchOffsetPairs();
		for (int i = 0; i < matchOffsetPairs.size(); i++) {
			pdsCond = new ExtendedList();
			pdsConfigTo = new PDSConfig(pdsClass.getMethodName(methodIndex) 
				+ String.valueOf(lookupSwitchInst.getOffset() 
					+ ((MatchOffsetPair) matchOffsetPairs.get(i)).getOffset()));
			addDefaultGlobalCond();
			addDefaultLocalCond(methodIndex);
			pdsCond.add("sv0=" + String.valueOf(((MatchOffsetPair) 
				matchOffsetPairs.get(i)).getMatch()));
			addPopStackCond(methodIndex);
			pdsRuleList.add(new PDSRule(pdsConfigFrom, pdsConfigTo, pdsCond));
		}
		pdsCond = new ExtendedList();
		pdsConfigTo = new PDSConfig(pdsClass.getMethodName(methodIndex) 
			+ String.valueOf(lookupSwitchInst.getOffset() 
				+ lookupSwitchInst.getDefaultOffset()));
		addDefaultGlobalCond();
		addDefaultLocalCond(methodIndex);
		for (int i = 0; i < matchOffsetPairs.size(); i++) {
			pdsCond.add("sv0!=" + String.valueOf(((MatchOffsetPair) 
				matchOffsetPairs.get(i)).getMatch()));
		}
		addPopStackCond(methodIndex);
		pdsRuleList.add(new PDSRule(pdsConfigFrom, pdsConfigTo, pdsCond));
	}
	
	private void ruleIreturn(AbstractInstruction inst, 
													 int methodIndex, 
													 int instIndex) {
		PDSConfig pdsConfigFrom;
		PDSConfig pdsConfigTo;
		
		pdsCond = new ExtendedList();
		pdsConfigFrom = new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instIndex));
		pdsConfigTo = new PDSConfig("");
		addStoreGlobalCond(PDSConstant.RETURN_VARIABLE);
		pdsRuleList.add(new PDSRule(pdsConfigFrom, pdsConfigTo,
			inst.getOpcode(), pdsCond));
	}
	
	private void ruleReturn(AbstractInstruction inst, 
													int methodIndex, 
													int instIndex) {
		PDSConfig pdsConfigFrom;
		PDSConfig pdsConfigTo;
		
		pdsCond = new ExtendedList();
		pdsConfigFrom = new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instIndex));
		if (pdsClass.getMethodName(methodIndex).equals(
				pdsClass.getClassName() + PDSConstant.INIT_STATIC_METHOD_NAME)) {
			pdsConfigTo = new PDSConfig(((PDSClass) pdsClassList.get(
				0)).getClassName() + PDSConstant.DEFAULT_START_STACK_SYMBOL);
			PDSClass includedPDSClass;
			for (int i = (getPDSClassListIndex(pdsClass.getClassName()) + 1); 
					i < pdsClassList.size(); i++) {
				includedPDSClass = (PDSClass) pdsClassList.get(i);
				if (includedPDSClass.isClinit()) {
					pdsConfigTo = new PDSConfig(includedPDSClass.getClassName() 
						+ PDSConstant.INIT_STATIC_METHOD_NAME + "0");
					break;
				}
			}
		}
		else {
			pdsConfigTo = new PDSConfig("");
		}
		addDefaultGlobalCond();
		pdsRuleList.add(new PDSRule(pdsConfigFrom, pdsConfigTo, 
			inst.getOpcode(), pdsCond));
	}
	
	private void ruleGetstatic(AbstractInstruction inst, 
														 int methodIndex, 
														 int instIndex) throws Exception {
		PDSConfig pdsConfigFrom;
		PDSConfig pdsConfigTo;
		ImmediateShortInstruction immShortInst;
		ConstantFieldrefInfo cpInfo;
		String fieldName;
		String formattedFieldName = null;
		boolean fakePush = false;
		boolean zeroPush = false;
		StringBuffer pdsLabel = new StringBuffer();
															 
		pdsCond = new ExtendedList();
		immShortInst = (ImmediateShortInstruction) inst;
		cpInfo = (ConstantFieldrefInfo) classFile.getConstantPoolEntry(
			immShortInst.getImmediateShort(), ConstantFieldrefInfo.class);
		fieldName = classFile.getConstantPoolEntryName(
			immShortInst.getImmediateShort());
		
		if (fieldName.equals(PDSConstant.JAVA_LANG_SYSTEM_OUT)
				|| (fieldName.indexOf(PDSConstant.CLASS_DOLLARSIGN) != -1)
				|| fieldName.equals(PDSConstant.JAVA_LANG_SYSTEM_ERR)) {
			fakePush = true;
		} else if (fieldName.endsWith(PDSConstant.ASSERTIONS_DISABLED)) {
			zeroPush = true;
		} else {
			formattedFieldName = PDSField.formatFieldName(fieldName);
			PDSClass currentPDSClass = getPDSClass(cpInfo.getClassInfo().getName(
				).replace('/', '_'));
			while (!pdsGlobal.isInPDSFieldList(formattedFieldName)) {
				currentPDSClass = getSuperPDSClass(currentPDSClass);
				formattedFieldName = currentPDSClass.getClassName() + "_" 
					+ cpInfo.getNameAndTypeInfo().getName();
			}
		}
		
		pdsConfigFrom = new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instIndex));
		pdsConfigTo = new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instIndex + 1));
			
		pdsLabel.append(immShortInst.getOpcode());
		pdsLabel.append(PDSConstant.ONE_SPACE);
		pdsLabel.append(immShortInst.getImmediateShort());
		if (!fakePush) {
			pdsLabel.append(PDSConstant.ONE_SPACE);
			pdsLabel.append(fieldName);
		}
			
		addDefaultGlobalCond();
		addDefaultLocalCond(methodIndex);
		if (fakePush) {
			addFakePushStackCond(methodIndex);
		} else if (zeroPush) {
			addPushStackCond(methodIndex, "0");
		} else {
			addPushStackCond(methodIndex, formattedFieldName);
		}
		pdsRuleList.add(new PDSRule(pdsConfigFrom, pdsConfigTo, 
			pdsLabel.toString(), pdsCond));
	}
	
	private void rulePutstatic(AbstractInstruction inst, 
														 int methodIndex, 
														 int instIndex) throws Exception {
		PDSConfig pdsConfigFrom;
		PDSConfig pdsConfigTo;
		ImmediateShortInstruction immShortInst;
		ConstantFieldrefInfo cpInfo;
		String fieldName;
		String formattedFieldName = null;
		PDSClass currentPDSClass;
		boolean isFakePut = false;
		StringBuffer pdsLabel = new StringBuffer();
		
		pdsCond = new ExtendedList();
		immShortInst = (ImmediateShortInstruction) inst;
		cpInfo = (ConstantFieldrefInfo) classFile.getConstantPoolEntry(
			immShortInst.getImmediateShort(), ConstantFieldrefInfo.class);
		fieldName = classFile.getConstantPoolEntryName(
			immShortInst.getImmediateShort());
		
		if (!fieldName.endsWith(PDSConstant.ASSERTIONS_DISABLED)
				&& (fieldName.indexOf(PDSConstant.CLASS_DOLLARSIGN) == -1)) {
			formattedFieldName = PDSField.formatFieldName(fieldName);
			currentPDSClass = getPDSClass(cpInfo.getClassInfo().getName().replace(
				'/', '_'));
			while (!pdsGlobal.isInPDSFieldList(formattedFieldName)) {
				currentPDSClass = getSuperPDSClass(currentPDSClass);
				formattedFieldName = currentPDSClass.getClassName() + "_" 
					+ cpInfo.getNameAndTypeInfo().getName();
			}
		} else {
			isFakePut = true;
		}
		pdsConfigFrom = new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instIndex));
		pdsConfigTo = new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instIndex + 1));
		
		pdsLabel.append(immShortInst.getOpcode());
		pdsLabel.append(PDSConstant.ONE_SPACE);
		pdsLabel.append(immShortInst.getImmediateShort());
		if (!isFakePut) {
			pdsLabel.append(PDSConstant.ONE_SPACE);
			pdsLabel.append(fieldName);
		}
		
		if (!isFakePut) {
			addStoreGlobalCond(formattedFieldName);
		}
		addDefaultLocalCond(methodIndex);
		addPopStackCond(methodIndex);
		pdsRuleList.add(new PDSRule(pdsConfigFrom, pdsConfigTo, 
			pdsLabel.toString(), pdsCond));
	}
	
	private void ruleGetfield(AbstractInstruction inst, 
														int methodIndex, 
														int instIndex) throws Exception {
		PDSConfig pdsConfigFrom;
		PDSConfig pdsConfigTo;
		ImmediateShortInstruction immShortInst;
		ConstantFieldrefInfo cpInfo;
		String fieldName;
		PDSClass currentPDSClass;
		StringBuffer pdsLabel = new StringBuffer();
		
		pdsCond = new ExtendedList();
		immShortInst = (ImmediateShortInstruction) inst;
		cpInfo = (ConstantFieldrefInfo) classFile.getConstantPoolEntry(
			immShortInst.getImmediateShort(), ConstantFieldrefInfo.class);
		fieldName = PDSField.formatFieldName(classFile.getConstantPoolEntryName(
			immShortInst.getImmediateShort()));
		currentPDSClass = getPDSClass(cpInfo.getClassInfo().getName().replace(
			'/', '_'));
		while (!pdsGlobal.isInPDSFieldList(fieldName)) {
			currentPDSClass = getSuperPDSClass(currentPDSClass);
			fieldName = currentPDSClass.getClassName() + "_" 
				+ cpInfo.getNameAndTypeInfo().getName();
		}
		pdsConfigFrom = new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instIndex));
		pdsConfigTo = new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instIndex + 1));
			
		pdsLabel.append(immShortInst.getOpcode());
		pdsLabel.append(PDSConstant.ONE_SPACE);
		pdsLabel.append(immShortInst.getImmediateShort());
		pdsLabel.append(PDSConstant.ONE_SPACE);
		pdsLabel.append(classFile.getConstantPoolEntryName(
			immShortInst.getImmediateShort()));
		pdsLabel.append(PDSConstant.ONE_SPACE);
		pdsLabel.append(cpInfo.getNameAndTypeInfo().getDescriptor());
			
		addDefaultGlobalCond();
		addDefaultLocalCond(methodIndex);
		pdsCond.add("sv0'=" + fieldName + "[sv0]");
		addDefaultStackCond(methodIndex, 1, false);
		pdsRuleList.add(new PDSRule(pdsConfigFrom, pdsConfigTo, 
			pdsLabel.toString(), pdsCond));
	}
	
	private void rulePutfield(AbstractInstruction inst, 
														int methodIndex, 
														int instIndex) throws Exception {
		PDSConfig pdsConfigFrom;
		PDSConfig pdsConfigTo;
		ImmediateShortInstruction immShortInst;
		ConstantFieldrefInfo cpInfo;
		String fieldName;
		PDSClass currentPDSClass;
		boolean isStaticField;
		StringBuffer pdsLabel = new StringBuffer();
		
		pdsCond = new ExtendedList();
		immShortInst = (ImmediateShortInstruction) inst;
		cpInfo = (ConstantFieldrefInfo) classFile.getConstantPoolEntry(
			immShortInst.getImmediateShort(), ConstantFieldrefInfo.class);
		fieldName = PDSField.formatFieldName(
			classFile.getConstantPoolEntryName(immShortInst.getImmediateShort()));
		currentPDSClass = getPDSClass(cpInfo.getClassInfo().getName().replace(
			'/', '_'));
		while (!pdsGlobal.isInPDSFieldList(fieldName)) {
			currentPDSClass = getSuperPDSClass(currentPDSClass);
			fieldName = currentPDSClass.getClassName() + "_" 
				+ cpInfo.getNameAndTypeInfo().getName();
		}
		pdsConfigFrom = new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instIndex));
		pdsConfigTo = new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instIndex + 1));
		
		pdsLabel.append(immShortInst.getOpcode());
		pdsLabel.append(PDSConstant.ONE_SPACE);
		pdsLabel.append(immShortInst.getImmediateShort());
		pdsLabel.append(PDSConstant.ONE_SPACE);
		pdsLabel.append(classFile.getConstantPoolEntryName(
			immShortInst.getImmediateShort()));
		
		isStaticField = addStoreGlobalCond(fieldName);
		addDefaultLocalCond(methodIndex);
		if (isStaticField)
			addPopStackCond(methodIndex);
		else
			addPopStackCond(methodIndex, 2, 2, false);
		pdsRuleList.add(new PDSRule(pdsConfigFrom, pdsConfigTo, 
			pdsLabel.toString(), pdsCond));
	}
	
	private void ruleInvokevirtual(AbstractInstruction inst, 
																 int methodIndex, 
																 int instIndex) throws Exception {
		PDSConfig pdsConfigFrom;
		PDSConfig pdsConfigTo;
		ImmediateShortInstruction immShortInst;
		ConstantMethodrefInfo cpInfo;
		String calleeClass;
		String calleeMethod;
		int calleeIndex;
		int parameterCount;
		int returnCount;
		String afterCallStackSymbol;
		StringBuffer pdsLabel = new StringBuffer();
		
		pdsCond = new ExtendedList();
		immShortInst = (ImmediateShortInstruction) inst;
		cpInfo = (ConstantMethodrefInfo) classFile.getConstantPoolEntry(
			immShortInst.getImmediateShort(), ConstantMethodrefInfo.class);
		calleeClass = cpInfo.getClassInfo().getName().replace('/', '_');
		calleeMethod = PDSClass.getPDSMethodSignature(calleeClass + "_" 
			+ cpInfo.getNameAndTypeInfo().getVerbose());
		PDSClass calleePDSClass = getPDSClass(calleeClass);
		
		pdsLabel.append(immShortInst.getOpcode());
		pdsLabel.append(PDSConstant.ONE_SPACE);
		pdsLabel.append(immShortInst.getImmediateShort());
		pdsLabel.append(PDSConstant.ONE_SPACE);
		pdsLabel.append(cpInfo.getClassInfo().getName());
		pdsLabel.append(PDSConstant.DOT);
		pdsLabel.append(cpInfo.getNameAndTypeInfo().getName());
		pdsLabel.append(PDSConstant.ONE_SPACE);
		pdsLabel.append(cpInfo.getNameAndTypeInfo().getDescriptor());
		
		/* Fake method call. */
		if ((calleePDSClass == null) 
				|| (calleePDSClass.getMethodIndex(calleeMethod) == -1)) {
			parameterCount = PDSClass.getParameterCount(
				cpInfo.getNameAndTypeInfo().getDescriptor()) + 1;
			returnCount = (PDSClass.isVoid(calleeMethod)) ? 0 : 1;
			if (parameterCount == returnCount) {
				ruleDefault(inst, methodIndex, instIndex, 1, pdsLabel.toString());
			} else if (parameterCount > returnCount) {
				rulePop(inst, methodIndex, instIndex, 
					parameterCount-returnCount, parameterCount-returnCount,
					pdsLabel.toString());
			} else {
				rulePush(inst, methodIndex, instIndex, pdsLabel.toString());
			}
			return;
		}
		
		while (!calleePDSClass.containsMethod(calleeMethod)) {
			calleePDSClass = getSuperPDSClass(calleePDSClass);
			calleeMethod = PDSClass.getPDSMethodSignature(
				calleePDSClass.getClassName() + "_" 
				+ cpInfo.getNameAndTypeInfo().getVerbose());
		}
		if (PDSConstant.verbose > 2) {
			System.out.println("      calleePDSClass=" 
				+ calleePDSClass.getClassName() + ", calleeMethod=" + calleeMethod);
		}
		calleeIndex = calleePDSClass.getMethodIndex(calleeMethod);
		parameterCount = calleePDSClass.getParameterCount(calleeIndex);
		pdsConfigFrom = new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instIndex));
		
		if (calleePDSClass.isVoid(calleeIndex)) {
			afterCallStackSymbol = pdsClass.getStackSymbol(
				methodIndex, instIndex + 1);
		} else {
			afterCallStackSymbol = pdsClass.getStackSymbol(
				methodIndex, instIndex + 1) + "a";
		}
		pdsConfigTo = new PDSConfig(PDSConstant.DEFAULT_CONTROL_LOCATION, 
			calleeMethod  + "0", afterCallStackSymbol);
			
		addDefaultGlobalCond();
		addDefaultLocalCond(methodIndex, true);
		addParameterPassingCond(parameterCount + 1);
		addPopStackCond(methodIndex, parameterCount + 1, parameterCount + 1, true);
		pdsRuleList.add(new PDSRule(pdsConfigFrom, pdsConfigTo, 
			pdsLabel.toString(), pdsCond));
		
		if (!calleePDSClass.isVoid(calleeIndex)) {
			rulePushReturn(methodIndex, instIndex, afterCallStackSymbol);
		}
		
		ruleExceptionHandler(methodIndex, instIndex, afterCallStackSymbol);
	}
	
	private void ruleInvokespecial(AbstractInstruction inst, 
																 int methodIndex, 
																 int instIndex) throws Exception {
		ImmediateShortInstruction immShortInst;
		ConstantMethodrefInfo cpInfo;
		String calleeClass;
		String calleeMethod;
		int calleeIndex;
		int parameterCount;
		int returnCount;
		PDSConfig pdsConfigFrom;
		PDSConfig pdsConfigTo;
		String afterCallStackSymbol;
		StringBuffer pdsLabel = new StringBuffer();
		
		pdsCond = new ExtendedList();
		immShortInst = (ImmediateShortInstruction) inst;
		cpInfo = (ConstantMethodrefInfo) classFile.getConstantPoolEntry(
			immShortInst.getImmediateShort(), ConstantMethodrefInfo.class);
		calleeClass = cpInfo.getClassInfo().getName().replace('/', '_');
		calleeMethod = PDSClass.getPDSMethodSignature(calleeClass + "_" 
			+ cpInfo.getNameAndTypeInfo().getVerbose());
		PDSClass calleePDSClass = getPDSClass(calleeClass);
		
		pdsLabel.append(immShortInst.getOpcode());
		pdsLabel.append(PDSConstant.ONE_SPACE);
		pdsLabel.append(immShortInst.getImmediateShort());
		pdsLabel.append(PDSConstant.ONE_SPACE);
		pdsLabel.append(cpInfo.getClassInfo().getName());
		pdsLabel.append(PDSConstant.DOT);
		pdsLabel.append(cpInfo.getNameAndTypeInfo().getName());
		pdsLabel.append(PDSConstant.ONE_SPACE);
		pdsLabel.append(cpInfo.getNameAndTypeInfo().getDescriptor());
		
		/* Handles invokespecial java/lang/AssertionError.<init> */
		if (calleeMethod.equals(PDSConstant.ASSERTION_ERROR_INIT)) {
			pdsConfigFrom = new PDSConfig(pdsClass.getStackSymbol(
				methodIndex, instIndex));
			pdsConfigTo = new PDSConfig(PDSConstant.ASSERTION_ERROR);

			addDefaultGlobalCond();
			pdsRuleList.add(new PDSRule(pdsConfigFrom, pdsConfigTo, 
				pdsLabel.toString(), pdsCond));
			assertIncluded = true;
			return;
		}
		
		/* Fake method call. */
		if ((calleePDSClass == null) 
				|| (calleePDSClass.getMethodIndex(calleeMethod) == -1)) {
			parameterCount = PDSClass.getParameterCount(
				cpInfo.getNameAndTypeInfo().getDescriptor()) + 1;
			returnCount = (PDSClass.isVoid(calleeMethod)) ? 0 : 1;
			if (parameterCount == returnCount) {
				ruleDefault(inst, methodIndex, instIndex, 1, pdsLabel.toString());
			} else if (parameterCount > returnCount) {
				rulePop(inst, methodIndex, instIndex, 
					parameterCount-returnCount, parameterCount-returnCount, 
					pdsLabel.toString());
			} else {
				rulePush(inst, methodIndex, instIndex, pdsLabel.toString());
			}
			return;
		}
		
		/* Could not find called method, find it from predecessor classes. */
		while (!calleePDSClass.containsMethod(calleeMethod)) {
			calleePDSClass = getSuperPDSClass(calleePDSClass);
			calleeMethod = PDSClass.getPDSMethodSignature(
				calleePDSClass.getClassName() + "_" 
				+ cpInfo.getNameAndTypeInfo().getVerbose());
		}
		
		calleeIndex = calleePDSClass.getMethodIndex(calleeMethod);
		parameterCount = calleePDSClass.getParameterCount(calleeIndex);
		pdsConfigFrom = new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instIndex));
		
		if (calleePDSClass.isVoid(calleeIndex)) {
			afterCallStackSymbol = pdsClass.getStackSymbol(methodIndex, 
				instIndex + 1);
		} else {
			afterCallStackSymbol = pdsClass.getStackSymbol(methodIndex, 
				instIndex + 1) + "a";
		}
		pdsConfigTo = new PDSConfig(PDSConstant.DEFAULT_CONTROL_LOCATION, 
			calleeMethod  + "0", afterCallStackSymbol);
		
		addDefaultGlobalCond();
		addDefaultLocalCond(methodIndex, true);
		addParameterPassingCond(parameterCount + 1);
		addPopStackCond(methodIndex, parameterCount + 1, parameterCount + 1, true);
		pdsRuleList.add(new PDSRule(pdsConfigFrom, pdsConfigTo,
			pdsLabel.toString(), pdsCond));
		
		/* An additional rule for returning value. */
		if (!calleePDSClass.isVoid(calleeIndex))
			rulePushReturn(methodIndex, instIndex, afterCallStackSymbol);
		
		ruleExceptionHandler(methodIndex, instIndex, afterCallStackSymbol);
	}
	
	private void ruleInvokestatic(AbstractInstruction inst, 
																int methodIndex, 
																int instIndex) throws Exception {
		ImmediateShortInstruction immShortInst;
		ConstantMethodrefInfo cpInfo;
		String calledClass;
		String calledMethod;
		PDSClass calledPDSClass;
		int calledIndex;
		int paramCount;
		int retCount;
		PDSConfig pdsConfigFrom;
		PDSConfig pdsConfigTo;
		String afterCallStackSymbol;
		StringBuffer pdsLabel = new StringBuffer();
		
		pdsCond = new ExtendedList();
		immShortInst = (ImmediateShortInstruction) inst;
		cpInfo = (ConstantMethodrefInfo) classFile.getConstantPoolEntry(
			immShortInst.getImmediateShort(), ConstantMethodrefInfo.class);
		calledClass = cpInfo.getClassInfo().getName().replace('/', '_');
		calledMethod = PDSClass.getPDSMethodSignature(calledClass + "_" 
			+ cpInfo.getNameAndTypeInfo().getVerbose());
		calledPDSClass = getPDSClass(calledClass);
		
		pdsLabel.append(immShortInst.getOpcode());
		pdsLabel.append(PDSConstant.ONE_SPACE);
		pdsLabel.append(immShortInst.getImmediateShort());
		pdsLabel.append(PDSConstant.ONE_SPACE);
		pdsLabel.append(cpInfo.getClassInfo().getName());
		pdsLabel.append(PDSConstant.DOT);
		pdsLabel.append(cpInfo.getNameAndTypeInfo().getName());
		pdsLabel.append(PDSConstant.ONE_SPACE);
		pdsLabel.append(cpInfo.getNameAndTypeInfo().getDescriptor());
		
		/* Called class not found. */
		if ((calledPDSClass == null) 
				|| (calledPDSClass.getMethodIndex(calledMethod) == -1)) {
			paramCount = PDSClass.getParameterCount(
				cpInfo.getNameAndTypeInfo().getDescriptor());
			retCount = (PDSClass.isVoid(calledMethod)) ? 0 : 1;
			if (paramCount == retCount) {
				ruleDefault(inst, methodIndex, instIndex, 1, pdsLabel.toString());
			} else if (paramCount > retCount) {
				rulePop(inst, methodIndex, instIndex, 
					paramCount-retCount, paramCount-retCount, pdsLabel.toString());
			} else {
				rulePush(inst, methodIndex, instIndex, pdsLabel.toString());
			}
			return;
		}
		
		calledIndex = calledPDSClass.getMethodIndex(calledMethod);
		paramCount = calledPDSClass.getParameterCount(calledIndex);
		pdsConfigFrom =  new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instIndex));
		
		if (calledPDSClass.isVoid(calledIndex)) {
			afterCallStackSymbol = pdsClass.getStackSymbol(methodIndex, 
				instIndex + 1);
		} else {
			afterCallStackSymbol = pdsClass.getStackSymbol(methodIndex, 
				instIndex + 1) + "a";
		}
		pdsConfigTo = new PDSConfig(PDSConstant.DEFAULT_CONTROL_LOCATION, 
			calledMethod  + "0", afterCallStackSymbol);
		
		
		
		addDefaultGlobalCond();
		addDefaultLocalCond(methodIndex, true);
		if (paramCount == 0) {
			addDefaultStackCond(methodIndex, 0, true);
		} else {
			addParameterPassingCond(paramCount);
			addPopStackCond(methodIndex, paramCount, paramCount, true);
		}
		pdsRuleList.add(new PDSRule(pdsConfigFrom, pdsConfigTo, 
			pdsLabel.toString(), pdsCond));
		
		if (!calledPDSClass.isVoid(calledIndex)) {
			rulePushReturn(methodIndex, instIndex, afterCallStackSymbol);
		}
		
		ruleExceptionHandler(methodIndex, instIndex, afterCallStackSymbol);
	}
	
	private void ruleInvokeinterface(AbstractInstruction inst, 
																	 int methodIndex, 
																	 int instIndex) throws Exception {
		PDSConfig pdsConfigFrom;
		PDSConfig pdsConfigTo;
		InvokeInterfaceInstruction invokeInfInst;
		ConstantInterfaceMethodrefInfo cpInfo;
		String calleeClass;
		String calleeMethod;
		PDSClass calleePDSClass;
		int parameterCount;
		int returnCount;
		int calleeIndex;
		String afterCallStackSymbol;
		StringBuffer pdsLabel = new StringBuffer();
		
		pdsCond = new ExtendedList();
		invokeInfInst = (InvokeInterfaceInstruction) inst;
		cpInfo = (ConstantInterfaceMethodrefInfo) classFile.getConstantPoolEntry(
			invokeInfInst.getImmediateShort(), ConstantInterfaceMethodrefInfo.class);
		calleeClass = cpInfo.getClassInfo().getName().replace('/', '_');
		calleeMethod = PDSClass.getPDSMethodSignature(calleeClass + "_" 
			+ cpInfo.getNameAndTypeInfo().getVerbose());
			
		pdsLabel.append(invokeInfInst.getOpcode());
		pdsLabel.append(PDSConstant.ONE_SPACE);
		pdsLabel.append(invokeInfInst.getImmediateShort());
		pdsLabel.append(PDSConstant.ONE_SPACE);
		pdsLabel.append(cpInfo.getClassInfo().getName());
		pdsLabel.append(PDSConstant.DOT);
		pdsLabel.append(cpInfo.getNameAndTypeInfo().getName());
		pdsLabel.append(PDSConstant.ONE_SPACE);
		pdsLabel.append(cpInfo.getNameAndTypeInfo().getDescriptor());
		
		calleePDSClass = getPDSClass(calleeClass);
		if ((calleePDSClass == null) 
				|| (calleePDSClass.getMethodIndex(calleeMethod) == -1)) {
			parameterCount = PDSClass.getParameterCount(
				cpInfo.getNameAndTypeInfo().getDescriptor()) + 1;
			returnCount = (PDSClass.isVoid(calleeMethod)) ? 0 : 1;
			if (parameterCount == returnCount) {
				ruleDefault(inst, methodIndex, instIndex, 1, pdsLabel.toString());
			} else if (parameterCount > returnCount) {
				rulePop(inst, methodIndex, instIndex, 
					parameterCount-returnCount, parameterCount-returnCount);
			} else {
				rulePush(inst, methodIndex, instIndex);
			}
			return;
		}
		
		calleeIndex = calleePDSClass.getMethodIndex(calleeMethod);
		parameterCount = calleePDSClass.getParameterCount(calleeIndex);
		pdsConfigFrom = new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instIndex));
		
		if (calleePDSClass.isVoid(calleeIndex)) {
			afterCallStackSymbol =  pdsClass.getStackSymbol(
				methodIndex, instIndex + 1);
		} else {
			afterCallStackSymbol =  pdsClass.getStackSymbol(
				methodIndex, instIndex + 1) + "a";
		}
		pdsConfigTo = new PDSConfig(PDSConstant.DEFAULT_CONTROL_LOCATION, 
			calleeMethod  + "0", afterCallStackSymbol);
			
		addDefaultGlobalCond();
		addDefaultLocalCond(methodIndex, true);
		addParameterPassingCond(parameterCount + 1);
		addPopStackCond(methodIndex, parameterCount + 1, parameterCount + 1, true);
		pdsRuleList.add(new PDSRule(pdsConfigFrom, pdsConfigTo, 
			pdsLabel.toString(), pdsCond));
		
		if (!calleePDSClass.isVoid(calleeIndex))
			rulePushReturn(methodIndex, instIndex, afterCallStackSymbol);
		
		ruleExceptionHandler(methodIndex, instIndex, afterCallStackSymbol);
	}
	
	private void rulePushReturn(int methodIndex, 
															int instIndex, 
															String afterCallStackSymbol) {
		PDSConfig pdsConfigFrom;
		PDSConfig pdsConfigTo;
		
		pdsCond = new ExtendedList();
		pdsConfigFrom = new PDSConfig(afterCallStackSymbol);
		pdsConfigTo = new PDSConfig( pdsClass.getStackSymbol(
			methodIndex, instIndex + 1));
		addDefaultGlobalCond();
		addDefaultLocalCond(methodIndex);
		addPushStackCond(methodIndex, PDSConstant.RETURN_VARIABLE);
		pdsRuleList.add(new PDSRule(pdsConfigFrom, pdsConfigTo, 0, pdsCond));
		pdsClass.addStackSymbol(methodIndex, afterCallStackSymbol);
	}
	
	private void ruleExceptionHandler(int methodIndex, 
																		int instIndex, 
																		String afterCallStackSymbol) {
		PDSConfig pdsConfigFrom;
		PDSConfig pdsConfigTo;
		int handlerPc;
		
		pdsCond = new ExtendedList();
		handlerPc = pdsClass.getHandlerPc(methodIndex, 
			pdsClass.getOffset(methodIndex, instIndex));
		pdsConfigFrom = new PDSConfig(
			PDSConstant.DEFAULT_EXCEPTION_CONTROL_LOCATION, afterCallStackSymbol);
		if (handlerPc == -1) {
			pdsConfigTo = new PDSConfig(
				PDSConstant.DEFAULT_EXCEPTION_CONTROL_LOCATION, "");
			addStoreGlobalCond(PDSConstant.RETURN_VARIABLE);
		}
		else {
			pdsConfigTo = new PDSConfig(pdsClass.getMethodName(methodIndex) 
				+ handlerPc);
			addDefaultGlobalCond();
			addDefaultLocalCond(methodIndex);
			addPushStackCond(methodIndex, PDSConstant.RETURN_VARIABLE);
		}
		pdsRuleList.add(new PDSRule(pdsConfigFrom, pdsConfigTo, 0, pdsCond));
	}
	
	private void ruleNew(AbstractInstruction inst, 
											 int methodIndex, 
											 int instIndex) throws Exception {
		ImmediateShortInstruction immShortInst;
		ConstantClassInfo constantClassInfo;
		String newClassName;
		PDSConfig pdsConfigFrom;
		PDSConfig pdsConfigTo;
		PDSClass newPDSClass;
		StringBuffer pdsLabel = new StringBuffer();
												 
		pdsCond = new ExtendedList();
		immShortInst = (ImmediateShortInstruction) inst;
		constantClassInfo = (ConstantClassInfo) classFile.getConstantPoolEntry(
			immShortInst.getImmediateShort(), ConstantClassInfo.class);
		newClassName = constantClassInfo.getName().replace('/', '_');
		pdsConfigFrom = new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instIndex));
		pdsConfigTo = new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instIndex + 1));
			
		pdsLabel.append(immShortInst.getOpcode());
		pdsLabel.append(PDSConstant.ONE_SPACE);
		pdsLabel.append(immShortInst.getImmediateShort());
		/*pdsLabel.append(PDSConstant.ONE_SPACE);
		pdsLabel.append(constantClassInfo.getName());
		pdsLabel.append(PDSConstant.ONE_SPACE);
		pdsLabel.append(constantClassInfo.getName());*/
			
		addDefaultGlobalCond();
		addDefaultLocalCond(methodIndex);
		newPDSClass = getPDSClass(newClassName);
		if (newPDSClass != null) {
			addPushStackCond(methodIndex, newPDSClass.getInstanceId() 
				+ newPDSClass.getInitUsed());
			newPDSClass.incInitUsed();
			propagateIncInitUsed(getSuperPDSClass(newPDSClass));
		}
		else {
			addPushStackCond(methodIndex, 1);
		}
		
		pdsRuleList.add(new PDSRule(pdsConfigFrom, pdsConfigTo, 
			pdsLabel.toString(), pdsCond));
	}
	
	private void propagateIncInitUsed(PDSClass superPDSClass) {
		if ((superPDSClass == null) 
				|| superPDSClass.getClassName().equals("java_lang_Object"))
			return;
		
		superPDSClass.incInitUsed();
		propagateIncInitUsed(getSuperPDSClass(superPDSClass));
	}
	
	private void ruleNewarray(AbstractInstruction inst, 
														int methodIndex, 
														int instIndex, 
														int newarrayCount) throws Exception {
		PDSConfig pdsConfigFrom;
		PDSConfig pdsConfigTo;
		
		pdsCond = new ExtendedList();
		pdsConfigFrom = new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instIndex));
		pdsConfigTo = new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instIndex + 1));
		addDefaultGlobalCond();
		addDefaultLocalCond(methodIndex);
		addPushStackCond(methodIndex, newarrayCount);
		pdsRuleList.add(new PDSRule(pdsConfigFrom, pdsConfigTo,
			inst.getOpcode(), pdsCond));
	}
	
	private void ruleArraylength(AbstractInstruction inst, 
															 int methodIndex, 
															 int instIndex) throws Exception {
		PDSConfig pdsConfigFrom;
		PDSConfig pdsConfigTo;
		int globalArrayCount;
		
		pdsCond = new ExtendedList();
		pdsConfigFrom = new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instIndex ));
		pdsConfigTo = new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instIndex + 1));
		addDefaultGlobalCond();
		addDefaultLocalCond(methodIndex);
		globalArrayCount = pdsGlobal.getArrayCount();
		for (int i = 0; i < globalArrayCount; i++) {
			pdsCond.add("!(sv0=" + String.valueOf(i+1) + ") | (sv0'=" 
				+ pdsGlobal.getArraySize(i) + ")");
		}
		for (int i = 0; i < pdsClass.getArraySizeCount(methodIndex); i++) {
			pdsCond.add("!(sv0=" + String.valueOf(globalArrayCount + i+1) 
				+ ") | (sv0'=" + pdsClass.getArraySize(methodIndex, i) + ")");
		}
		addDefaultStackCond(methodIndex, 1, false);
		pdsRuleList.add(new PDSRule(pdsConfigFrom, pdsConfigTo, 
			inst.getOpcode(), pdsCond));
	}
	
	private void ruleAthrow(AbstractInstruction inst, 
													int methodIndex, 
													int instIndex) throws Exception {
		PDSConfig pdsConfigFrom;
		PDSConfig pdsConfigTo;
		int handlerPc;
		
		pdsCond = new ExtendedList();
		pdsConfigFrom = new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instIndex));
		pdsConfigTo = new PDSConfig(
			PDSConstant.DEFAULT_EXCEPTION_CONTROL_LOCATION, "");
		handlerPc = pdsClass.getHandlerPc(methodIndex, 
			pdsClass.getOffset(methodIndex, instIndex));
		if (handlerPc > -1) {
			addDefaultGlobalCond();
			pdsConfigTo = new PDSConfig(pdsClass.getMethodName(methodIndex) 
				+ handlerPc);
			addDefaultLocalCond(methodIndex);
			addDefaultStackCond(methodIndex);
		}
		else {	//ret'=sv0
			addStoreGlobalCond(PDSConstant.RETURN_VARIABLE);
		}
		pdsRuleList.add(new PDSRule(pdsConfigFrom, pdsConfigTo, 
			inst.getOpcode(), pdsCond));
	}
	
	private void ruleCheckcast(AbstractInstruction inst, 
														 int methodIndex, 
														 int instIndex) throws Exception {
		ImmediateShortInstruction immShortInst;
		ConstantClassInfo constantClassInfo;
		String className;
		PDSClass checkingPDSClass;
		PDSConfig pdsConfigFrom;
		PDSConfig pdsConfigTo;
		StringBuffer pdsLabel = new StringBuffer();
		
		pdsCond = new ExtendedList();
		immShortInst = (ImmediateShortInstruction) inst;
		constantClassInfo = (ConstantClassInfo) classFile.getConstantPoolEntry(
			immShortInst.getImmediateShort(), ConstantClassInfo.class);
		className = constantClassInfo.getName().replace('/', '_');
		checkingPDSClass = getPDSClass(className);
		if (checkingPDSClass == null) {
			ruleDefault(inst, methodIndex, instIndex);
			return;
		}
		pdsConfigFrom = new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instIndex));
		pdsConfigTo = new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instIndex + 1));
			
		pdsLabel.append(immShortInst.getOpcode());
		pdsLabel.append(PDSConstant.ONE_SPACE);
		pdsLabel.append(immShortInst.getImmediateShort());
			
		addDefaultGlobalCond();
		addDefaultLocalCond(methodIndex);
		pdsCond.add("(sv0>=" + String.valueOf(checkingPDSClass.getInstanceId()) 
			+ ")" + " & (sv0=<" + String.valueOf(checkingPDSClass.getInstanceId() 
			+ checkingPDSClass.getInitCount() - 1) + ")");
		addDefaultStackCond(methodIndex);
		pdsRuleList.add(new PDSRule(pdsConfigFrom, pdsConfigTo, 
			pdsLabel.toString(), pdsCond));
		
		pdsCond = new ExtendedList();
		pdsConfigFrom = new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instIndex));
		pdsConfigTo = new PDSConfig(
			PDSConstant.DEFAULT_EXCEPTION_CONTROL_LOCATION, "");
		addDefaultGlobalCond();
		addDefaultLocalCond(methodIndex);
		pdsCond.add("(sv0<" + String.valueOf(checkingPDSClass.getInstanceId()) 
			+ ")" + " & (sv0>" + String.valueOf(checkingPDSClass.getInstanceId() 
			+ checkingPDSClass.getInitCount() - 1) + ")");
		addDefaultStackCond(methodIndex);
		pdsRuleList.add(new PDSRule(pdsConfigFrom, pdsConfigTo, 0, pdsCond));
		
	}
	
	private void ruleInstanceof(AbstractInstruction inst, 
															int methodIndex, 
															int instIndex) {
		PDSConfig pdsConfigFrom;
		PDSConfig pdsConfigTo;
		PDSClass pdsClassI;
		
		pdsCond = new ExtendedList();
		pdsConfigFrom = new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instIndex));
		pdsConfigTo = new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instIndex + 1));
		addDefaultGlobalCond();
		addDefaultLocalCond(methodIndex);

		for (int i = 0; i < pdsClassList.size(); i++) {
			pdsClassI = (PDSClass) pdsClassList.get(i);
			if (pdsClassI.getInitCount() > 0) {
				pdsCond.add("!((sv0>=" + String.valueOf(pdsClassI.getInstanceId()) 
					+ ")" + " & (sv0=<" + String.valueOf(pdsClassI.getInstanceId() 
						+ pdsClassI.getInitCount() - 1) + ")) | (sv0'=1)");
			}
		}
		addDefaultStackCond(methodIndex, 1, false);
		pdsRuleList.add(new PDSRule(pdsConfigFrom, pdsConfigTo, 
			inst.getOpcode(), pdsCond));
	}
	
	private int ruleMultianewarray(AbstractInstruction inst, 
																 int methodIndex, 
																 int instIndex, 
																 int newarrayCount) throws Exception {
		PDSConfig pdsConfigFrom;
		PDSConfig pdsConfigTo;
		int dim;
		int size;
		int l;
		int outerArray;
		int currentIndex;
		int m;
		
		pdsCond = new ExtendedList();
		pdsConfigFrom = new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instIndex));
		pdsConfigTo = new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instIndex + 1));
		
		//initialize sub-arrays
		dim = ((MultianewarrayInstruction) inst).getDimensions();
		if (pdsClass.getMethodName(methodIndex).equals(
				pdsClass.getClassName() + PDSConstant.INIT_STATIC_METHOD_NAME)) {
			addDefaultGlobalFieldCond();
			for (int i = 0; i < newarrayCount-1; i++) {
				addDefaultGlobalArrayCond(i);
			}
			addDefaultLocalCond(methodIndex);
			addPushStackCond(methodIndex, newarrayCount);
			
			size = pdsGlobal.getArraySize(newarrayCount - 1);
			l = dim - 1;
			outerArray = 1;
			currentIndex = newarrayCount;
			while (l > 0) {
				l--;
				for (m = 0; m < outerArray; m++) {
					size = pdsGlobal.getArraySize(currentIndex + m - 1);
					for (int n = 0; n < size; n++) {
						newarrayCount++;
						pdsCond.add("arr" + String.valueOf(currentIndex + m) 
							+ "'[" + n + "]=" + String.valueOf(newarrayCount));
					}
				}
				currentIndex++;
				outerArray *= size;
			}
			
			for (int i = currentIndex-1; i < pdsGlobal.getArrayCount(); i++) {
				addDefaultGlobalArrayCond(i);
			}
		}
		else {
			addDefaultGlobalCond();
			addDefaultLocalCond(methodIndex);
			addPushStackCond(methodIndex, pdsGlobal.getArrayCount() + newarrayCount);
			
			size = pdsClass.getArraySize(methodIndex, newarrayCount - 1);
			l = dim - 1;
			outerArray = 1;
			currentIndex = newarrayCount;
			while (l > 0) {
				l--;
				for (m = 0; m < outerArray; m++) {
					size = pdsClass.getArraySize(methodIndex, currentIndex + m - 1);
					for (int n = 0; n < size; n++) {
						newarrayCount++;
						pdsCond.add("arr" + String.valueOf(pdsGlobal.getArrayCount() 
								+ currentIndex + m) + "'[" + n + "]=" 
							+ String.valueOf(pdsGlobal.getArrayCount() + newarrayCount));
					}
				}
				currentIndex++;
				outerArray *= size;
			}
		}
		
		pdsRuleList.add(new PDSRule(pdsConfigFrom, pdsConfigTo, 
			inst.getOpcode(), pdsCond));
		return newarrayCount;
	}
	
	private void ruleDefault(AbstractInstruction inst, 
													 int methodIndex, 
													 int instIndex) {
		ruleDefault(inst, methodIndex, instIndex, 0, null);
	}
	
	private void ruleDefault(AbstractInstruction inst, 
													 int methodIndex, 
													 int instIndex, 
													 int fromStackIndex) {
		ruleDefault(inst, methodIndex, instIndex, fromStackIndex, null);
	}
	
	private void ruleDefault(AbstractInstruction abstractInstruction, 
													 int methodIndex, 
													 int instructionIndex, 
													 int fromStackIndex, 
													 String pdsLabel) {
		PDSConfig pdsConfigFrom = new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instructionIndex));
		PDSConfig pdsConfigTo = new PDSConfig(pdsClass.getStackSymbol(
			methodIndex, instructionIndex + 1));
			
		pdsCond = new ExtendedList();
		addDefaultGlobalCond();
		addDefaultLocalCond(methodIndex);
		addDefaultStackCond(methodIndex, fromStackIndex, false);
		pdsRuleList.add(new PDSRule(pdsConfigFrom, pdsConfigTo, 
			pdsLabel, pdsCond));
	}
	
	private void addDefaultGlobalCond() {
		addDefaultGlobalFieldCond();
		addDefaultGlobalArrayCond();
	}
	
	private void addDefaultGlobalFieldCond() {
		for (int i = 0; i < pdsGlobal.getFieldCount(); i++)
			addDefaultGlobalFieldCond(i);
	}
	
	private void addDefaultGlobalFieldCond(int fieldIndex) {
		PDSField pdsField = (PDSField) pdsGlobal.get(fieldIndex);
		if (pdsField.isStatic())
			pdsCond.add(pdsField.getFieldName() + "'=" 
				+ pdsField.getFieldName());
		else
			pdsCond.add("A i (" + pdsField.getFromIndex() 
				+ "," + pdsField.getToIndex() + ") " 
				+ pdsField.getFieldName() + "'[i]=" 
				+ pdsField.getFieldName() + "[i]");
				
	}
	
	private void addDefaultGlobalArrayCond() {
		for (int i = 0; i < pdsGlobal.getArrayCount(); i++)
			addDefaultGlobalArrayCond(i);
	}
	
	private void addDefaultGlobalArrayCond(int arrayIndex) {
		PDSArray pdsArray = (PDSArray) pdsGlobal.getPDSArray(arrayIndex);
		pdsCond.add("A i (0," + String.valueOf(pdsArray.getArraySize() - 1) 
			+ ") arr" + String.valueOf(arrayIndex + 1) + "'[i]=arr" 
			+ String.valueOf(arrayIndex + 1) + "[i]");
	}
	
	private boolean addStoreGlobalCond(String var) {
		boolean isStaticField = true;
		PDSField pdsField;
		for (int i = 0; i < pdsGlobal.getFieldCount(); i++) {
			pdsField = (PDSField) pdsGlobal.get(i);
			if (var.equals(pdsField.getFieldName())) {
				if (pdsField.isStatic()) {
					pdsCond.add(pdsField.getFieldName() + "'=sv0");
				}
				else {
					pdsCond.add(pdsField.getFieldName() + "'[sv1]=sv0");
					pdsCond.add("A i (" + pdsField.getFromIndex() 
						+ "," + pdsField.getToIndex() + ") (!(i!=sv1) | (" 
						+ pdsField.getFieldName() + "'[i]=" 
						+ pdsField.getFieldName() + "[i]))");
					isStaticField = false;
				}
			}
			else
				addDefaultGlobalFieldCond(i);
		}
		addDefaultGlobalArrayCond();
		
		return isStaticField;
	}
	
	private void addStoreArrayCond(int methodIndex) {
		addDefaultGlobalFieldCond();
		String arrayIndex;
		ExtendedList invariant;
		PDSArray pdsArray;
		for (int i = 0; i < pdsGlobal.getArrayCount(); i++) {
			invariant = new ExtendedList();
			for (int j = 0; j < pdsGlobal.getArrayCount(); j++) {
				if (i != j) {
					arrayIndex = String.valueOf(j + 1);
					pdsArray = (PDSArray) pdsGlobal.getPDSArray(j);
					invariant.add("A i (0," + String.valueOf(pdsArray.getArraySize() - 1) 
						+ ") arr" + arrayIndex + "'[i]=arr" + arrayIndex + "[i]");
				}
			}
			for (int j = 0; j < pdsClass.getArraySizeCount(methodIndex); j++) {
				arrayIndex = String.valueOf(pdsGlobal.getArrayCount() + j + 1);
				invariant.add("A i (0," + String.valueOf(pdsClass.getArraySize(methodIndex, j) - 1) 
					+ ") arr" + arrayIndex + "'[i]=arr" + arrayIndex + "[i]");
			}
			arrayIndex = String.valueOf(i + 1);
			if (invariant.isEmpty()) {
				pdsCond.add("!(sv2=" + arrayIndex + ") | ((arr" + arrayIndex + "'[sv1]=sv0) & (A i (0," 
					+ String.valueOf(pdsGlobal.getArraySize(i) - 1) + ") (!(i!=sv1) | (arr"
					+ arrayIndex + "'[i]=arr" + arrayIndex + "[i]))))");
			}
			else
				pdsCond.add("!(sv2=" + arrayIndex + ") | ((arr" + arrayIndex + "'[sv1]=sv0) & (A i (0," 
					+ String.valueOf(pdsGlobal.getArraySize(i) - 1) + ") (!(i!=sv1) | (arr"
					+ arrayIndex + "'[i]=arr" + arrayIndex + "[i]))) & " + invariant.printBooleanExpression() + ")");
		}
		
		for (int i = 0; i < pdsClass.getMaxLocal(methodIndex); i++)
			pdsCond.add("var" + i + "'=" + "var" + i);
		for (int i = 0; i < pdsClass.getArraySizeCount(methodIndex); i++) {
			
			invariant = new ExtendedList();
			for (int j = 0; j < pdsGlobal.getArrayCount(); j++) {
				arrayIndex = String.valueOf(j + 1);
				pdsArray = (PDSArray) pdsGlobal.getPDSArray(j);
				invariant.add("A i (0," + String.valueOf(pdsArray.getArraySize() - 1) 
					+ ") arr" + arrayIndex + "'[i]=arr" + arrayIndex + "[i]");
			}
			for (int j = 0; j < pdsClass.getArraySizeCount(methodIndex); j++) {
				if (i != j) {
					arrayIndex = String.valueOf(pdsGlobal.getArrayCount() + j + 1);
					invariant.add("A i (0," + String.valueOf(pdsClass.getArraySize(methodIndex, j) - 1) 
						+ ") arr" + arrayIndex + "'[i]=arr" + arrayIndex + "[i]");
				}
			}
			arrayIndex = String.valueOf(pdsGlobal.getArrayCount() + i + 1);
			if (invariant.isEmpty())
				pdsCond.add("!(sv2=" + arrayIndex + ") | ((arr" + arrayIndex + "'[sv1]=sv0) & (A i (0," 
					+ String.valueOf(pdsClass.getArraySize(methodIndex, i) - 1) + ") (!(i!=sv1) | (arr" 
					+ arrayIndex + "'[i]=arr" + arrayIndex + "[i]))))");
			else
				pdsCond.add("!(sv2=" + arrayIndex + ") | ((arr" + arrayIndex + "'[sv1]=sv0) & (A i (0," 
					+ String.valueOf(pdsClass.getArraySize(methodIndex, i) - 1) + ") (!(i!=sv1) | (arr" 
					+ arrayIndex + "'[i]=arr" + arrayIndex + "[i]))) & " + invariant.printBooleanExpression() + ")");
		}
	}
	
	private void addDefaultLocalCond(int methodIndex) {
		addDefaultLocalCond(methodIndex, false);
	}
	
	private void addDefaultLocalCond(int methodIndex, boolean isDoubleQuote) {
		String middleString;
		if (isDoubleQuote) middleString = "''";
		else middleString = "'";
		for (int i = 0; i < pdsClass.getMaxLocal(methodIndex); i++)
			pdsCond.add("var" + i + middleString + "=var" + i);
		for (int i = 0; i < pdsClass.getArraySizeCount(methodIndex); i++)
			pdsCond.add("A i (0," + String.valueOf(pdsClass.getArraySize(methodIndex, i) - 1) 
				+ ") arr" + String.valueOf(pdsGlobal.getArrayCount() + i + 1) + middleString 
				+ "[i]=arr" + String.valueOf(pdsGlobal.getArrayCount() + i + 1) + "[i]");
	}
	
	private void addStoreLocalCond(int methodIndex, int varIndex) {
		addStoreLocalCond(methodIndex, varIndex, "sv0");
	}
	
	private void addStoreLocalCond(int methodIndex, int varIndex, String item) {
		for (int i = 0; i < pdsClass.getMaxLocal(methodIndex); i++) {
			if (i == varIndex)
				pdsCond.add("var" + i + "'=" + item);
			else
				pdsCond.add("var" + i + "'=var" + i);
		}
		for (int i = 0; i < pdsClass.getArraySizeCount(methodIndex); i++)
			pdsCond.add("A i (0," + String.valueOf(pdsClass.getArraySize(methodIndex, i) - 1) 
				+ ") arr" + String.valueOf(pdsGlobal.getArrayCount() + i + 1)
				+ "'[i]=arr" + String.valueOf(pdsGlobal.getArrayCount() + i + 1) + "[i]");
	}
	
	private void addDefaultStackCond(int methodIndex) {
		addDefaultStackCond(methodIndex, 0, false);
	}
	
	private void addDefaultStackCond(int methodIndex, int from, boolean isDoubleQuote) {
		String middleString;
		if (isDoubleQuote) middleString = "''=";
		else middleString = "'=";
		for (int i = from; i < pdsClass.getMaxStack(methodIndex); i++)
			pdsCond.add("sv" + i + middleString + "sv" + i);
	}
	
	private void addPushStackCond(int methodIndex, int item) {
		pdsCond.add("sv0'=" + String.valueOf(item));
		for (int i = 1; i < pdsClass.getMaxStack(methodIndex); i++)
			pdsCond.add("sv" + i + "'=sv" + String.valueOf(i - 1));
	}
	
	private void addPushStackCond(int methodIndex, String item) {
		pdsCond.add("sv0'=" + item);
		for (int i = 1; i < pdsClass.getMaxStack(methodIndex); i++)
			pdsCond.add("sv" + i + "'=sv" + String.valueOf(i - 1));
	}
	
	private void addFakePushStackCond(int methodIndex) {
		addFakePushStackCond(methodIndex, 1);
	}
	
	private void addFakePushStackCond(int methodIndex, int from) {
		for (int i = from; i < pdsClass.getMaxStack(methodIndex); i++)
			pdsCond.add("sv" + i + "'=sv" + String.valueOf(i - 1));
	}
	
	private void addPopStackCond(int methodIndex) {
		addPopStackCond(methodIndex, 1, 1, false);
	}
	
	private void addPopStackCond(int methodIndex, int from, int step, boolean isDoubleQuote) {
		String middleString;
		if (isDoubleQuote) middleString = "''=";
		else middleString = "'=";
		for (int i = from; i < pdsClass.getMaxStack(methodIndex); i++)
			pdsCond.add("sv" + String.valueOf(i - step) + middleString + "sv" + i);
	}
	
	private void addParameterPassingCond(int paramCount) {
		for (int i = 0; i < paramCount; i++)
			pdsCond.add("var" + String.valueOf(paramCount - i - 1) + "'=sv" + i);
	}
	
	private PDSClass getPDSClass(String className) {
		for (int i = 0; i < pdsClassList.size(); i++) {
			if (((PDSClass) pdsClassList.get(i)).getClassName().equals(className))
				return (PDSClass) pdsClassList.get(i);
		}
		return null;
	}
	
	private PDSClass getSuperPDSClass(PDSClass subPDSClass) {
		String superClassName = subPDSClass.getSuperClassName();
		for (int i = 0; i < pdsClassList.size(); i++)
			if (((PDSClass) pdsClassList.get(i)).getClassName().equals(superClassName))
				return (PDSClass) pdsClassList.get(i);
		return null;
	}
	
	private int getPDSClassListIndex(String className) {
		for (int i = 0; i < pdsClassList.size(); i++)
			if (((PDSClass) pdsClassList.get(i)).getClassName().equals(className))
				return i;
		return -1;
	}
	
}
