/*
    Copyright (C) 2004 Dejvuth Suwimonteerabuth

    This file is part of jMoped.

    jMoped is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    jMoped is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with jMoped; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package jmoped;

import org.gjt.jclasslib.io.*;
import org.gjt.jclasslib.structures.*;

import java.util.ArrayList;
import java.io.*;
import java.util.jar.*;
import java.net.URL;

/**
 * Structure of the pushdown system.
 */
public class PDSStructure {
	/*
	 * Private Method:
	 *   void setPDSIncludedClass
	 *   void addPDSIncludedClass
	 *   void setPDSIncludedClassFromJar
	 *   void addPDSIncludedClassFromJar
	 *   void setPDSClassInitCount
	 *   void setSuperPDSClassInitCount
	 *   void setPDSGlobal
	 *   void setPDSLocalList
	 *   void addPDSLocalList
	 *   void setInitPDSConfig
	 *   void setPDSRuleList
	 *   boolean isExistInPDSClassList
	 *   PDSClass getSuperPDSClass
	 */
	
	private ArrayList pdsClassList;
	private ArrayList classFileList;
	private ArrayList includedPDSBasicMethodList;
	private PDSGlobal pdsGlobal;
	private ExtendedList pdsLocalList;
	private PDSConfig pdsInitConfig;
	private PDSRuleList pdsRuleList;
	
	/**
	 * Initializes the structure.
	 *
	 * @param fileName the name of class file.
	 * @param bits the number of bits of variables.
	 */
	public PDSStructure(String fileName, int bits) throws Exception {
		this(fileName, bits, PDSConstant.DEFAULT_CONTROL_LOCATION);
	}
	
	/**
	 * Initializes the structure.
	 *
	 * @param fileName the name of class file.
	 * @param bits the number of bits of variables.
	 * @param initControlLocation the initial control location.
	 */
	public PDSStructure(String fileName, 
											int bits, 
											String initControlLocation) throws Exception {
		File file = new File(fileName);
		
		ClassFile classFile = ClassFileReader.readFromFile(file);
		classFileList = new ArrayList();
		classFileList.add(classFile);
		
		PDSClass pdsClass = new PDSClass(classFile, bits);
		pdsClassList = new ArrayList();
		pdsClassList.add(pdsClass);
		
		includedPDSBasicMethodList = pdsClass.getIncludedPDSBasicMethodList();
		
		setPDSIncludedClass(file, pdsClass, bits);
		setPDSGlobal(bits);
		setPDSLocalList(pdsClassList, bits);
		setInitPDSConfig(initControlLocation, pdsClass.getClassName() 
			+ PDSConstant.DEFAULT_START_STACK_SYMBOL);
		setPDSRuleList(classFile, pdsClass, classFileList, pdsClassList
			, pdsGlobal);
	}
	
	/**
	 * Initializes the structure from the jar file.
	 *
	 * @param jarFileName the jar file name.
	 * @param jarEntryName the class entry in the jar file.
	 * @param bits the number of bits of variables.
	 */
	public PDSStructure(String jarFileName, 
											String jarEntryName, 
											int bits) throws Exception {
		this(jarFileName, jarEntryName, bits
			, PDSConstant.DEFAULT_CONTROL_LOCATION);
	}
	
	/**
	 * Initializes the structure from the jar file.
	 *
	 * @param jarFileName the jar file name.
	 * @param jarEntryName the class entry in the jar file.
	 * @param bits the number of bits of variables.
	 * @param initControlLocation the initial control location.
	 */
	public PDSStructure(String jarFileName, 
											String jarEntryName, 
											int bits, 
											String initControlLocation) throws Exception {
		JarFile jarFile = new JarFile(jarFileName);
		JarEntry jarEntry = jarFile.getJarEntry(jarEntryName);
		
		ClassFile classFile = ClassFileReader.readFromInputStream(
			jarFile.getInputStream(jarEntry));
		classFileList = new ArrayList();
		classFileList.add(classFile);
		
		PDSClass pdsClass = new PDSClass(classFile, bits);
		pdsClassList = new ArrayList();
		pdsClassList.add(pdsClass);
		
		includedPDSBasicMethodList = pdsClass.getIncludedPDSBasicMethodList();
		
		setPDSIncludedClassFromJar(jarFile, pdsClass, bits);
		setPDSGlobal(bits);
		setPDSLocalList(pdsClassList, bits);
		setInitPDSConfig(initControlLocation, pdsClass.getClassName() 
			+ PDSConstant.DEFAULT_START_STACK_SYMBOL);
		setPDSRuleList(classFile, pdsClass, classFileList, pdsClassList, 
			pdsGlobal);
	}
	
	/**
	 *
	 */
	public ArrayList getClassFileList() {
		return classFileList;
	}
	
	public ArrayList getPDSClassList() {
		return pdsClassList;
	}
	
	private void setPDSIncludedClass(File mainFile, 
																	 PDSClass pdsCallerClass, 
																	 int bits) throws Exception {
		String includedClassName;
		File includedFile;
		ClassFile includedClassFile;
		PDSClass pdsCalleeClass;
		ArrayList includedClassNameList;
		
		includedClassNameList = pdsCallerClass.getIncludedClassNameList();
		for (int i = 0; i < includedClassNameList.size(); i++) {
			if (isExistInPDSClassList((String) includedClassNameList.get(i))) {
				continue;
			}
			includedClassName = ((String) includedClassNameList.get(i)).replaceAll(
					"_", File.separator) + ".class";
			if (mainFile.getParent() != null) {
				includedClassName = mainFile.getParent() + File.separator 
					+ includedClassName;
			}
			
			includedFile = new File(includedClassName);
			if (!includedFile.exists()) {
				URL includedClassURL = this.getClass().getResource(includedClassName);
				setPDSIncludedClassFromJar(new JarFile(includedClassURL.getFile(
					).substring(5, includedClassURL.getFile().indexOf('!')))
						, pdsCallerClass, bits);
				continue;
			}
				
			includedClassFile = ClassFileReader.readFromFile(includedFile);
			classFileList.add(includedClassFile);
			pdsCalleeClass = new PDSClass(includedClassFile, 
				includedPDSBasicMethodList, bits);
			includedPDSBasicMethodList 
				= pdsCalleeClass.getIncludedPDSBasicMethodList();
			pdsClassList.add(pdsCalleeClass);
			addPDSIncludedClass(mainFile, pdsCallerClass, pdsCalleeClass, bits);
		}
		setPDSClassInitCount();
	}
	
	private void addPDSIncludedClass(File mainFile, 
																	 PDSClass pdsCallerClass, 
																	 PDSClass pdsCalleeClass, 
																	 int bits) throws Exception {
		if (pdsCallerClass.getClassName().equals(pdsCalleeClass.getClassName()))
			return;
		
		String includedClassName;
		File includedFile;
		ClassFile includedClassFile;
		PDSClass pdsCalleeOfCalleeClass;
		ArrayList includedClassNameList;
		
		includedClassNameList = pdsCalleeClass.getIncludedClassNameList();
		for (int i = 0; i < includedClassNameList.size(); i++) {
			if (isExistInPDSClassList((String) includedClassNameList.get(i)))
				continue;
			includedClassName = ((String) includedClassNameList.get(i)).replaceAll(
				"_", File.separator) + ".class";
					
			if (mainFile.getParent() != null) {
				includedClassName = mainFile.getParent() + File.separator 
					+ includedClassName;
			}
					
			includedFile = new File(includedClassName);
			if (!includedFile.exists()) {
				URL includedClassURL = this.getClass().getResource(includedClassName);
				addPDSIncludedClassFromJar(new JarFile(includedClassURL.getFile(
					).substring(5, includedClassURL.getFile().indexOf('!')))
						, pdsCallerClass, pdsCalleeClass, bits);
				continue;
			}
			
			includedClassFile = ClassFileReader.readFromFile(includedFile);
			classFileList.add(includedClassFile);
			pdsCalleeOfCalleeClass = new PDSClass(includedClassFile, 
				includedPDSBasicMethodList, bits);
			includedPDSBasicMethodList 
				= pdsCalleeOfCalleeClass.getIncludedPDSBasicMethodList();
			pdsClassList.add(pdsCalleeOfCalleeClass);
			addPDSIncludedClass(mainFile, pdsCalleeClass, pdsCalleeOfCalleeClass, 
				bits);
		}
	}
	
	private void setPDSIncludedClassFromJar(JarFile jarFile, 
																					PDSClass pdsCallerClass, 
																					int bits) throws Exception {
		String includedClassName;
		JarEntry jarEntry;
		ClassFile includedClassFile;
		PDSClass pdsCalleeClass;
		ArrayList includedClassNameList;
		
		includedClassNameList = pdsCallerClass.getIncludedClassNameList();
		for (int i = 0; i < includedClassNameList.size(); i++) {
			includedClassName = (String) includedClassNameList.get(i);
			if (isExistInPDSClassList(includedClassName))
				continue;
			jarEntry = jarFile.getJarEntry(includedClassName.replace('_', '/') 
				+ ".class");
			includedClassFile = ClassFileReader.readFromInputStream(
				jarFile.getInputStream(jarEntry));
			classFileList.add(includedClassFile);
			pdsCalleeClass = new PDSClass(includedClassFile, 
				includedPDSBasicMethodList, bits);
			includedPDSBasicMethodList 
				= pdsCalleeClass.getIncludedPDSBasicMethodList();
			pdsClassList.add(pdsCalleeClass);
			addPDSIncludedClassFromJar(jarFile, pdsCallerClass, pdsCalleeClass, 
				bits);
		}
		setPDSClassInitCount();
	}
	
	private void addPDSIncludedClassFromJar(JarFile jarFile, 
																					PDSClass pdsCallerClass, 
																					PDSClass pdsCalleeClass, 
																					int bits) throws Exception {
		if (pdsCallerClass.getClassName().equals(pdsCalleeClass.getClassName()))
			return;
		
		String includedClassName;
		JarEntry jarEntry;
		ClassFile includedClassFile;
		PDSClass pdsCalleeOfCalleeClass;
		
		ArrayList includedClassNameList = pdsCalleeClass.getIncludedClassNameList();
		for (int i = 0; i < includedClassNameList.size(); i++) {
			includedClassName = (String) includedClassNameList.get(i);
			if (isExistInPDSClassList(includedClassName)) {
				continue;
			}
			jarEntry = jarFile.getJarEntry(includedClassName.replace('_', '/') 
				+ ".class");
			includedClassFile = ClassFileReader.readFromInputStream(
				jarFile.getInputStream(jarEntry));
			classFileList.add(includedClassFile);
			pdsCalleeOfCalleeClass = new PDSClass(includedClassFile, 
				includedPDSBasicMethodList, bits);
			includedPDSBasicMethodList 
				= pdsCalleeOfCalleeClass.getIncludedPDSBasicMethodList();
			pdsClassList.add(pdsCalleeOfCalleeClass);
			addPDSIncludedClassFromJar(jarFile, pdsCalleeClass, 
				pdsCalleeOfCalleeClass, bits);
		}
	}
	
	
	
	private void setPDSClassInitCount() {
		ArrayList newClassNameList;
		PDSClass pdsClass, includedPDSClass;
		for (int k = 0; k < pdsClassList.size(); k++) {
			pdsClass = (PDSClass) pdsClassList.get(k);
			newClassNameList = pdsClass.getNewClassNameList();
			for (int i = 0; i < newClassNameList.size(); i++) {
				if (!newClassNameList.get(i).equals(pdsClass.getClassName())) {
					for (int j = 0; j < pdsClassList.size(); j++) {
						includedPDSClass = (PDSClass) pdsClassList.get(j);
						if (newClassNameList.get(i).equals(
								includedPDSClass.getClassName())) {
							includedPDSClass.incInitCount();
							break;
						}
					}
				} else {
					pdsClass.incInitCount();
				}
			}
			setSuperPDSClassInitCount(getSuperPDSClass(pdsClass), 
				pdsClass.getInitCount());
		}
	}
	
	private void setSuperPDSClassInitCount(PDSClass superPDSClass, int count) {
		if (superPDSClass == null)
			return;
		
		superPDSClass.incInitCount(count);
		setSuperPDSClassInitCount(getSuperPDSClass(superPDSClass), count);
	}
	
	private void setPDSGlobal(int bits) throws Exception {
		int instanceId = 1;
		PDSClass pdsClass;
		String superClassName;
		
		pdsGlobal = new PDSGlobal(bits);
		for (int i = 0; i < pdsClassList.size(); i++) {
			pdsClass = (PDSClass) pdsClassList.get(i);
			pdsGlobal.addGlobal(pdsClass.getPDSGlobal(), instanceId, 
				pdsClass.getInitCount());
			pdsClass.setInstanceId(instanceId);
			superClassName = pdsClass.getSuperClassName();
			if ((i != (pdsClassList.size() - 1)) 
					&& !((PDSClass) pdsClassList.get(i+1)).getClassName().equals(
						superClassName))
				instanceId += pdsClass.getInitCount();
		}
		
		for (int i = 0; i < pdsClassList.size(); i++) {
			if (!((PDSClass) pdsClassList.get(i)).isVoid()) {
				pdsGlobal.addReturnVariable();
				break;
			}
		}
	}
	
	private void setPDSLocalList(ArrayList pdsClassList, int bits) {
		pdsLocalList = new ExtendedList();
		for (int i = 0; i < pdsClassList.size(); i++)
			addPDSLocalList((PDSClass) pdsClassList.get(i), bits);
	}
	
	private void addPDSLocalList(PDSClass pdsClass, int bits) {
		int maxStack;
		int maxLocal;
		ArrayList arraySizeList;
		ExtendedList variableList;
		int i, j;
		
		for(i = 0; i < pdsClass.getMethodCount(); i++) {
			maxStack = pdsClass.getMaxStack(i);
			maxLocal = pdsClass.getMaxLocal(i);
			arraySizeList = pdsClass.getArraySizeList(i);
			if((maxStack + maxLocal + arraySizeList.size()) == 0)
				continue;
			variableList = new ExtendedList();
			for (j = 0; j < maxStack; j++)
				variableList.add("sv" + j);
			for (j = 0; j < maxLocal; j++)
				variableList.add("var" + j);
			for (j = 0; j < arraySizeList.size(); j++)
				variableList.add("arr" + String.valueOf(pdsGlobal.getArrayCount() 
					+ j + 1) + "[" + ((Integer) arraySizeList.get(j)).toString() + "]");
			pdsLocalList.add(new PDSLocal(pdsClass.getStackSymbolList(i), 'I', 
				variableList, bits));
		}
	}
	
	private void setInitPDSConfig(String initControlLocation, 
																String initStackAlphabet) {
		PDSClass pdsClass;
		for (int i = 0; i < pdsClassList.size(); i++) {
			pdsClass = (PDSClass) pdsClassList.get(i);
			if (pdsClass.isClinit()) {
				pdsInitConfig = new PDSConfig(initControlLocation
					, pdsClass.getClassName() + PDSConstant.INIT_STATIC_METHOD_NAME 
						+ "0");
				return;
			}
				
		}
		pdsInitConfig = new PDSConfig(initControlLocation, initStackAlphabet);
	}
	
	private void setPDSRuleList(ClassFile classFile, 
															PDSClass pdsClass, 
															ArrayList classFileList, 
															ArrayList pdsClassList, 
															PDSGlobal pdsGlobal) throws Exception{
		pdsRuleList = new PDSRuleList(classFile, pdsClass, pdsClassList, 
			pdsGlobal);
		for (int i = 1; i < pdsClassList.size(); i++) {
			pdsRuleList.addPDSRuleList(new PDSRuleList(
				(ClassFile) classFileList.get(i), 
				(PDSClass) pdsClassList.get(i), 
				pdsClassList, 
				pdsGlobal));
		}
	}
	
	private boolean isExistInPDSClassList(String className) {
		for (int i = 0; i < pdsClassList.size(); i++) {
			if (className.equals(((PDSClass) pdsClassList.get(i)).getClassName()))
				return true;
		}
		return false;
	}
	
	private PDSClass getSuperPDSClass(PDSClass subPDSClass) {
		String superClassName = subPDSClass.getSuperClassName();
		for (int i = 0; i < pdsClassList.size(); i++)
			if (((PDSClass) pdsClassList.get(i)).getClassName().equals(
					superClassName))
				return (PDSClass) pdsClassList.get(i);
		return null;
	}
	
	/**
	 * Prints the pushdown system in Moped's syntax.
	 *
	 * @return the pushdown system.
	 */
	public StringBuffer print() {
		StringBuffer output = new StringBuffer();
		
		if (pdsGlobal.size() > 0) {
			output.append(pdsGlobal.print());
			output.append(PDSConstant.NEW_LINE);
		}
		
		for(int i = 0; i < pdsLocalList.size(); i++) {
			output.append(((PDSLocal) pdsLocalList.get(i)).print());
			output.append(PDSConstant.NEW_LINE);
		}
		
		output.append(PDSConstant.NEW_LINE);
		output.append("(");
		output.append(pdsInitConfig.print());
		output.append(")");
		output.append(PDSConstant.NEW_LINE);
		output.append(PDSConstant.NEW_LINE);
		
		for(int i = 0; i < pdsRuleList.size(); i++) {
			output.append((pdsRuleList.get(i)).print());
			output.append(PDSConstant.NEW_LINE);
		}
		
		return output;
	}
	
	public StringBuffer printRemopla() {
		StringBuffer output = new StringBuffer();
		PDSRule pdsRule;
		String currentModule = null;
		String lhsStackSymbol;
		char lastChar;
		PDSLocal pdsLocal;
		
		if (pdsGlobal.size() > 0) {
			output.append(pdsGlobal.printRemopla());
			output.append(PDSConstant.NEW_LINE);
		}
		
		output.append(pdsInitConfig.printRemopla());
		output.append(PDSConstant.NEW_LINE);
		output.append(PDSConstant.NEW_LINE);
		
		for(int i = 0; i < pdsRuleList.size(); i++) {
			
			pdsRule = pdsRuleList.get(i);
			lhsStackSymbol = pdsRule.getLhsConfig().getFirstStackSymbol();
			if (currentModule == null 
					|| !lhsStackSymbol.startsWith(currentModule)) {
						
				if (currentModule != null) {
					output.append("}\n\n");
				}
				currentModule = lhsStackSymbol;
				lastChar = currentModule.charAt(currentModule.length() - 1);
				if (lastChar >= '0' && lastChar <= '9') {
					currentModule = currentModule.substring(0, 
						currentModule.length() - 1);
					lastChar = currentModule.charAt(currentModule.length() - 1);
				}
				output.append("module void ");
				output.append(lhsStackSymbol);
				output.append("() {\n");
				
				for(int j = 0; j < pdsLocalList.size(); j++) {
					pdsLocal = (PDSLocal) pdsLocalList.get(j);
					if (pdsLocal.getFirstStackSymbol().equals(lhsStackSymbol)) {
						output.append(pdsLocal.printRemopla());
						output.append(PDSConstant.NEW_LINE);
						break;
					}
				}
			}
			
			output.append(pdsRule.printRemopla());
			output.append(PDSConstant.NEW_LINE);
		}
		output.append("}\n");
		
		return output;
	}
	
}
